@@ -8,13 +8,11 @@
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2016 by Bradford W. Mott, Stephen Anthony
+// Copyright (c) 1995-2022 by Bradford W. Mott, Stephen Anthony
 // and the Stella Team
 //
 // See the file "License.txt" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
-//
-// $Id: MouseControl.cxx 3302 2016-04-02 23:47:46Z stephena $
 //============================================================================
 
 #include "Console.hxx"
@@ -26,10 +24,9 @@
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 MouseControl::MouseControl(Console& console, const string& mode)
-  : myProps(console.properties()),
-    myLeftController(console.leftController()),
-    myRightController(console.rightController()),
-    myCurrentModeNum(0)
+  : myProps{console.properties()},
+    myLeftController{console.leftController()},
+    myRightController{console.rightController()}
 {
   istringstream m_axis(mode);
   string m_mode;
@@ -37,118 +34,80 @@ MouseControl::MouseControl(Console& console, const string& mode)
 
   if(BSPF::equalsIgnoreCase(m_mode, "none"))
   {
-    myModeList.push_back(MouseMode("Mouse input is disabled"));
+    myModeList.emplace_back("Mouse input is disabled");
     return;
   }
   else if(!BSPF::equalsIgnoreCase(m_mode, "auto") && m_mode.length() == 2 &&
           m_mode[0] >= '0' && m_mode[0] <= '8' &&
           m_mode[1] >= '0' && m_mode[1] <= '8')
   {
-    Axis xaxis = Axis(int(m_mode[0]) - '0');
-    Axis yaxis = Axis(int(m_mode[1]) - '0');
+    const MouseControl::Type xaxis = static_cast<MouseControl::Type>(static_cast<int>(m_mode[0]) - '0');
+    const MouseControl::Type yaxis = static_cast<MouseControl::Type>(static_cast<int>(m_mode[1]) - '0');
     ostringstream msg;
-    msg << "Mouse X-axis is ";
-    Controller::Type xtype = Controller::Joystick, ytype = Controller::Joystick;
+    Controller::Type xtype = Controller::Type::Joystick, ytype = Controller::Type::Joystick;
     int xid = -1, yid = -1;
-    switch(xaxis)
-    {
-      case NoControl:
-        msg << "not used";
-        break;
-      case Paddle0:
-        xtype = Controller::Paddles;
-        xid = 0;
-        msg << "Paddle 0";
-        break;
-      case Paddle1:
-        xtype = Controller::Paddles;
-        xid = 1;
-        msg << "Paddle 1";
-        break;
-      case Paddle2:
-        xtype = Controller::Paddles;
-        xid = 2;
-        msg << "Paddle 2";
-        break;
-      case Paddle3:
-        xtype = Controller::Paddles;
-        xid = 3;
-        msg << "Paddle 3";
-        break;
-      case Driving0:
-        xtype = Controller::Driving;
-        xid = 0;
-        msg << "Driving 0";
-        break;
-      case Driving1:
-        xtype = Controller::Driving;
-        xid = 1;
-        msg << "Driving 1";
-        break;
-      case MindLink0:
-        xtype = Controller::MindLink;
-        xid = 0;
-        msg << "MindLink 0";
-        break;
-      case MindLink1:
-        xtype = Controller::MindLink;
-        xid = 1;
-        msg << "MindLink 1";
-        break;
-    }
+
+    const auto MControlToController = [&msg](MouseControl::Type axis,
+                                             Controller::Type& type, int& id) {
+      switch(axis)
+      {
+        case MouseControl::Type::NoControl:
+          msg << "not used";
+          break;
+        case MouseControl::Type::LeftPaddleA:
+          type = Controller::Type::Paddles;
+          id = 0;
+          msg << "Left Paddle A";
+          break;
+        case MouseControl::Type::LeftPaddleB:
+          type = Controller::Type::Paddles;
+          id = 1;
+          msg << "Left Paddle B";
+          break;
+        case MouseControl::Type::RightPaddleA:
+          type = Controller::Type::Paddles;
+          id = 2;
+          msg << "Right Paddle A";
+          break;
+        case MouseControl::Type::RightPaddleB:
+          type = Controller::Type::Paddles;
+          id = 3;
+          msg << "Right Paddle B";
+          break;
+        case MouseControl::Type::LeftDriving:
+          type = Controller::Type::Driving;
+          id = 0;
+          msg << "Left Driving";
+          break;
+        case MouseControl::Type::RightDriving:
+          type = Controller::Type::Driving;
+          id = 1;
+          msg << "Right Driving";
+          break;
+        case MouseControl::Type::LeftMindLink:
+          type = Controller::Type::MindLink;
+          id = 0;
+          msg << "Left MindLink";
+          break;
+        case MouseControl::Type::RightMindLink:
+          type = Controller::Type::MindLink;
+          id = 1;
+          msg << "Right MindLink";
+          break;
+      }
+    };
+
+    msg << "Mouse X-axis is ";
+    MControlToController(xaxis, xtype, xid);
     msg << ", Y-axis is ";
-    switch(yaxis)
-    {
-      case NoControl:
-        msg << "not used";
-        break;
-      case Paddle0:
-        ytype = Controller::Paddles;
-        yid = 0;
-        msg << "Paddle 0";
-        break;
-      case Paddle1:
-        ytype = Controller::Paddles;
-        yid = 1;
-        msg << "Paddle 1";
-        break;
-      case Paddle2:
-        ytype = Controller::Paddles;
-        yid = 2;
-        msg << "Paddle 2";
-        break;
-      case Paddle3:
-        ytype = Controller::Paddles;
-        yid = 3;
-        msg << "Paddle 3";
-        break;
-      case Driving0:
-        ytype = Controller::Driving;
-        yid = 0;
-        msg << "Driving 0";
-        break;
-      case Driving1:
-        ytype = Controller::Driving;
-        yid = 1;
-        msg << "Driving 1";
-        break;
-      case MindLink0:
-        ytype = Controller::MindLink;
-        yid = 0;
-        msg << "MindLink 0";
-        break;
-      case MindLink1:
-        ytype = Controller::MindLink;
-        yid = 1;
-        msg << "MindLink 1";
-        break;
-    }
-    myModeList.push_back(MouseMode(xtype, xid, ytype, yid, msg.str()));
+    MControlToController(yaxis, ytype, yid);
+
+    myModeList.emplace_back(xtype, xid, ytype, yid, msg.str());
   }
 
   // Now consider the possible modes for the mouse based on the left
   // and right controllers
-  bool noswap = BSPF::equalsIgnoreCase(myProps.get(Console_SwapPorts), "NO");
+  const bool noswap = BSPF::equalsIgnoreCase(myProps.get(PropType::Console_SwapPorts), "NO");
   if(noswap)
   {
     addLeftControllerModes(noswap);
@@ -165,11 +124,11 @@ MouseControl::MouseControl(Console& console, const string& mode)
   int m_range = 100;
   if(!(m_axis >> m_range))
     m_range = 100;
-  Paddles::setPaddleRange(m_range);
+  Paddles::setDigitalPaddleRange(m_range);
 
   // If the mouse isn't used at all, we still need one item in the list
   if(myModeList.size() == 0)
-    myModeList.push_back(MouseMode("Mouse not used for current controllers"));
+    myModeList.emplace_back("Mouse not used for current controllers");
 
 #if 0
   for(const auto& m: myModeList)
@@ -178,13 +137,18 @@ MouseControl::MouseControl(Console& console, const string& mode)
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-const string& MouseControl::next()
+const string& MouseControl::change(int direction)
 {
+  myCurrentModeNum = BSPF::clampw(myCurrentModeNum + direction, 0,
+                                  static_cast<int>(myModeList.size() - 1));
   const MouseMode& mode = myModeList[myCurrentModeNum];
-  myCurrentModeNum = (myCurrentModeNum + 1) % myModeList.size();
 
-  myLeftController.setMouseControl(mode.xtype, mode.xid, mode.ytype, mode.yid);
-  myRightController.setMouseControl(mode.xtype, mode.xid, mode.ytype, mode.yid);
+  const bool leftControl =
+    myLeftController.setMouseControl(mode.xtype, mode.xid, mode.ytype, mode.yid);
+  const bool rightControl =
+    myRightController.setMouseControl(mode.xtype, mode.xid, mode.ytype, mode.yid);
+
+  myHasMouseControl = leftControl || rightControl;
 
   return mode.message;
 }
@@ -194,7 +158,7 @@ void MouseControl::addLeftControllerModes(bool noswap)
 {
   if(controllerSupportsMouse(myLeftController))
   {
-    if(myLeftController.type() == Controller::Paddles)
+    if(myLeftController.type() == Controller::Type::Paddles)
     {
       if(noswap)  addPaddleModes(0, 1, 0, 1);
       else        addPaddleModes(2, 3, 0, 1);
@@ -205,7 +169,7 @@ void MouseControl::addLeftControllerModes(bool noswap)
       msg << "Mouse is left " << myLeftController.name() << " controller";
       Controller::Type type = myLeftController.type();
       int id = noswap ? 0 : 1;
-      myModeList.push_back(MouseMode(type, id, type, id, msg.str()));
+      myModeList.emplace_back(type, id, type, id, msg.str());
     }
   }
 }
@@ -215,7 +179,7 @@ void MouseControl::addRightControllerModes(bool noswap)
 {
   if(controllerSupportsMouse(myRightController))
   {
-    if(myRightController.type() == Controller::Paddles)
+    if(myRightController.type() == Controller::Type::Paddles)
     {
       if(noswap)  addPaddleModes(2, 3, 2, 3);
       else        addPaddleModes(0, 1, 2, 3);
@@ -226,7 +190,7 @@ void MouseControl::addRightControllerModes(bool noswap)
       msg << "Mouse is right " << myRightController.name() << " controller";
       Controller::Type type = myRightController.type();
       int id = noswap ? 1 : 0;
-      myModeList.push_back(MouseMode(type, id, type, id, msg.str()));
+      myModeList.emplace_back(type, id, type, id, msg.str());
     }
   }
 }
@@ -234,16 +198,16 @@ void MouseControl::addRightControllerModes(bool noswap)
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void MouseControl::addPaddleModes(int lport, int rport, int lname, int rname)
 {
-  Controller::Type type = Controller::Paddles;
+  const Controller::Type type = Controller::Type::Paddles;
   ostringstream msg;
   msg << "Mouse is Paddle " << lname << " controller";
-  MouseMode mode0(type, lport, type, lport, msg.str());
+  const MouseMode mode0(type, lport, type, lport, msg.str());
 
   msg.str("");
   msg << "Mouse is Paddle " << rname << " controller";
-  MouseMode mode1(type, rport, type, rport, msg.str());
+  const MouseMode mode1(type, rport, type, rport, msg.str());
 
-  if(BSPF::equalsIgnoreCase(myProps.get(Controller_SwapPaddles), "NO"))
+  if(BSPF::equalsIgnoreCase(myProps.get(PropType::Controller_SwapPaddles), "NO"))
   {
     myModeList.push_back(mode0);
     myModeList.push_back(mode1);
@@ -262,5 +226,5 @@ bool MouseControl::controllerSupportsMouse(Controller& controller)
   // We can pass in dummy values here, since the controllers will be
   // initialized by a call to next() once the system is up and running
   return controller.setMouseControl(
-      Controller::Joystick, -1, Controller::Joystick, -1);
+      Controller::Type::Joystick, -1, Controller::Type::Joystick, -1);
 }
