@@ -1,22 +1,23 @@
 //============================================================================
 //
-//   SSSS    tt          lll  lll       
-//  SS  SS   tt           ll   ll        
-//  SS     tttttt  eeee   ll   ll   aaaa 
+//   SSSS    tt          lll  lll
+//  SS  SS   tt           ll   ll
+//  SS     tttttt  eeee   ll   ll   aaaa
 //   SSSS    tt   ee  ee  ll   ll      aa
 //      SS   tt   eeeeee  ll   ll   aaaaa  --  "An Atari 2600 VCS Emulator"
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2016 by Bradford W. Mott, Stephen Anthony
+// Copyright (c) 1995-2022 by Bradford W. Mott, Stephen Anthony
 // and the Stella Team
 //
 // See the file "License.txt" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
-//
-// $Id: CartCTYWidget.cxx 3308 2016-05-24 16:55:45Z stephena $
 //============================================================================
 
+#include "OSystem.hxx"
+#include "Debugger.hxx"
+#include "CartDebug.hxx"
 #include "CartCTY.hxx"
 #include "PopUpWidget.hxx"
 #include "CartCTYWidget.hxx"
@@ -26,9 +27,9 @@ CartridgeCTYWidget::CartridgeCTYWidget(
       GuiObject* boss, const GUI::Font& lfont, const GUI::Font& nfont,
       int x, int y, int w, int h, CartridgeCTY& cart)
   : CartDebugWidget(boss, lfont, nfont, x, y, w, h),
-    myCart(cart)
+    myCart{cart}
 {
-  uInt16 size = 8 * 4096;
+  constexpr uInt16 size = 8 * 4096;
 
   string info =
     "Chetiry cartridge, eight 4K banks (bank 0 is ARM code and is ignored)\n"
@@ -36,21 +37,21 @@ CartridgeCTYWidget::CartridgeCTYWidget(
     "  $F040 - $F07F (R), $F000 - $F03F (W)\n"
     "\nTHIS SCHEME IS NOT FULLY IMPLEMENTED OR TESTED\n";
 
-  int xpos = 10,
-      ypos = addBaseInformation(size, "Chris D. Walton", info) + myLineHeight;
+  constexpr int xpos = 2;
+  const int ypos = addBaseInformation(size, "Chris D. Walton", info) + myLineHeight;
 
   VariantList items;
-  VarList::push_back(items, "1 ($FF5)");
-  VarList::push_back(items, "2 ($FF6)");
-  VarList::push_back(items, "3 ($FF7)");
-  VarList::push_back(items, "4 ($FF8)");
-  VarList::push_back(items, "5 ($FF9)");
-  VarList::push_back(items, "6 ($FFA)");
-  VarList::push_back(items, "7 ($FFB)");
+  VarList::push_back(items, "1 ($FFF5)");
+  VarList::push_back(items, "2 ($FFF6)");
+  VarList::push_back(items, "3 ($FFF7)");
+  VarList::push_back(items, "4 ($FFF8)");
+  VarList::push_back(items, "5 ($FFF9)");
+  VarList::push_back(items, "6 ($FFFA)");
+  VarList::push_back(items, "7 ($FFFB)");
   myBank =
-    new PopUpWidget(boss, _font, xpos, ypos-2, _font.getStringWidth("0 ($FFx) "),
-                    myLineHeight, items, "Set bank: ",
-                    _font.getStringWidth("Set bank: "), kBankChanged);
+    new PopUpWidget(boss, _font, xpos, ypos-2, _font.getStringWidth("0 ($FFFx)"),
+                    myLineHeight, items, "Set bank     ",
+                    0, kBankChanged);
   myBank->setTarget(this);
   addFocusWidget(myBank);
 }
@@ -59,17 +60,17 @@ CartridgeCTYWidget::CartridgeCTYWidget(
 void CartridgeCTYWidget::saveOldState()
 {
   myOldState.internalram.clear();
-  
-  for(uInt32 i = 0; i < this->internalRamSize();i++)
-  {
+
+  for(uInt32 i = 0; i < internalRamSize(); ++i)
     myOldState.internalram.push_back(myCart.myRAM[i]);
-  }
+
+  myOldState.bank = myCart.getBank();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void CartridgeCTYWidget::loadConfig()
 {
-  myBank->setSelectedIndex(myCart.getBank()-1);
+  myBank->setSelectedIndex(myCart.getBank()-1, myCart.getBank() != myOldState.bank);
 
   CartDebugWidget::loadConfig();
 }
@@ -80,9 +81,9 @@ void CartridgeCTYWidget::handleCommand(CommandSender* sender,
 {
   if(cmd == kBankChanged)
   {
-    myCart.unlockBank();
+    myCart.unlockHotspots();
     myCart.bank(myBank->getSelected()+1);
-    myCart.lockBank();
+    myCart.lockHotspots();
     invalidate();
   }
 }
@@ -92,17 +93,17 @@ string CartridgeCTYWidget::bankState()
 {
   ostringstream& buf = buffer();
 
-  static const char* spot[] = {
-    "", "$FF5", "$FF6", "$FF7", "$FF8", "$FF9", "$FFA", "$FFB"
+  static constexpr std::array<const char*, 8> spot = {
+    "", "$FFF5", "$FFF6", "$FFF7", "$FFF8", "$FFF9", "$FFFA", "$FFFB"
   };
-  uInt16 bank = myCart.getBank();
+  const uInt16 bank = myCart.getBank();
   buf << "Bank = " << std::dec << bank << ", hotspot = " << spot[bank];
 
   return buf.str();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-uInt32 CartridgeCTYWidget::internalRamSize() 
+uInt32 CartridgeCTYWidget::internalRamSize()
 {
   return 64;
 }
@@ -114,12 +115,12 @@ uInt32 CartridgeCTYWidget::internalRamRPort(int start)
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-string CartridgeCTYWidget::internalRamDescription() 
+string CartridgeCTYWidget::internalRamDescription()
 {
   ostringstream desc;
   desc << "$F000 - $F03F used for Write Access\n"
        << "$F040 - $F07F used for Read Access";
-  
+
   return desc.str();
 }
 
@@ -154,8 +155,8 @@ uInt8 CartridgeCTYWidget::internalRamGetValue(int addr)
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-string CartridgeCTYWidget::internalRamLabel(int addr) 
+string CartridgeCTYWidget::internalRamLabel(int addr)
 {
-  CartDebug& dbg = instance().debugger().cartDebug();
+  const CartDebug& dbg = instance().debugger().cartDebug();
   return dbg.getLabel(addr + 0xF040, false);
 }
