@@ -1,20 +1,18 @@
 //============================================================================
 //
-//   SSSS    tt          lll  lll       
-//  SS  SS   tt           ll   ll        
-//  SS     tttttt  eeee   ll   ll   aaaa 
+//   SSSS    tt          lll  lll
+//  SS  SS   tt           ll   ll
+//  SS     tttttt  eeee   ll   ll   aaaa
 //   SSSS    tt   ee  ee  ll   ll      aa
 //      SS   tt   eeeeee  ll   ll   aaaaa  --  "An Atari 2600 VCS Emulator"
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2016 by Bradford W. Mott, Stephen Anthony
+// Copyright (c) 1995-2022 by Bradford W. Mott, Stephen Anthony
 // and the Stella Team
 //
 // See the file "License.txt" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
-//
-// $Id: CartDPCWidget.cxx 3308 2016-05-24 16:55:45Z stephena $
 //============================================================================
 
 #include "CartDPC.hxx"
@@ -27,105 +25,105 @@ CartridgeDPCWidget::CartridgeDPCWidget(
       GuiObject* boss, const GUI::Font& lfont, const GUI::Font& nfont,
       int x, int y, int w, int h, CartridgeDPC& cart)
   : CartDebugWidget(boss, lfont, nfont, x, y, w, h),
-    myCart(cart)
+    myCart{cart}
 {
-  uInt16 size = cart.mySize;
-
+  constexpr int V_GAP = 4;
+  const size_t size = cart.mySize;
   ostringstream info;
-  info << "DPC cartridge, two 4K banks + 2K display bank\n"
-       << "DPC registers accessible @ $F000 - $F07F\n"
-       << "  $F000 - $F03F (R), $F040 - $F07F (W)\n"
 
-       << "Startup bank = " << cart.myStartBank << "\n";
+  info << "DPC cartridge, two 4K banks + 2K display bank\n"
+       << "DPC registers accessible @ $" << Common::Base::HEX4 << 0xF000 << " - $" << 0xF07F << "\n"
+       << "  $" << 0xF000 << " - " << 0xF03F << " (R), $" << 0xF040 << " - $" << 0xF07F << " (W)\n"
+
+       << "Startup bank = " << cart.startBank() << " or undetermined\n";
 
   // Eventually, we should query this from the debugger/disassembler
-  for(uInt32 i = 0, offset = 0xFFC, spot = 0xFF8; i < 2; ++i, offset += 0x1000)
+  constexpr uInt32 spot = 0xFF8;
+  for(uInt32 i = 0, offset = 0xFFC; i < 2; ++i, offset += 0x1000)
   {
     uInt16 start = (cart.myImage[offset+1] << 8) | cart.myImage[offset];
     start -= start % 0x1000;
     info << "Bank " << i << " @ $" << Common::Base::HEX4 << (start + 0x80) << " - "
-         << "$" << (start + 0xFFF) << " (hotspot = $" << (spot+i) << ")\n";
+         << "$" << (start + 0xFFF) << " (hotspot = $" << (0xF000 + spot + i) << ")\n";
   }
 
-  int xpos = 10,
+  int xpos = 2,
       ypos = addBaseInformation(size, "Activision (Pitfall II)", info.str()) +
               myLineHeight;
 
   VariantList items;
-  VarList::push_back(items, "0 ($FF8)");
-  VarList::push_back(items, "1 ($FF9)");
+  for(int bank = 0; bank < 2; ++bank)
+  {
+    ostringstream buf;
+
+    buf << "#" << std::dec << bank << " ($" << Common::Base::HEX4 << (0xFFF8 + bank) << ")";
+    VarList::push_back(items, buf.str());
+  }
+
   myBank =
-    new PopUpWidget(boss, _font, xpos, ypos-2, _font.getStringWidth("0 ($FFx) "),
-                    myLineHeight, items, "Set bank: ",
-                    _font.getStringWidth("Set bank: "), kBankChanged);
+    new PopUpWidget(boss, _font, xpos, ypos-2, _font.getStringWidth("#0 ($FFFx)"),
+                    myLineHeight, items, "Set bank     ",
+                    0, kBankChanged);
   myBank->setTarget(this);
   addFocusWidget(myBank);
-  ypos += myLineHeight + 8;
+  ypos += myLineHeight + V_GAP * 3;
 
   // Data fetchers
-  int lwidth = _font.getStringWidth("Data Fetchers: ");
-  new StaticTextWidget(boss, _font, xpos, ypos, lwidth,
-        myFontHeight, "Data Fetchers: ", kTextAlignLeft);
+  new StaticTextWidget(boss, _font, xpos, ypos, "Data fetchers ");
 
   // Top registers
-  lwidth = _font.getStringWidth("Counter Registers: ");
-  xpos = 18;  ypos += myLineHeight + 4;
-  new StaticTextWidget(boss, _font, xpos, ypos, lwidth,
-        myFontHeight, "Top Registers: ", kTextAlignLeft);
+  int lwidth = _font.getStringWidth("Counter registers ");
+  xpos = 2 + _font.getMaxCharWidth() * 2; ypos += myLineHeight + 4;
+  new StaticTextWidget(boss, _font, xpos, ypos, "Top registers ");
   xpos += lwidth;
 
-  myTops = new DataGridWidget(boss, _nfont, xpos, ypos-2, 8, 1, 2, 8, Common::Base::F_16);
+  myTops = new DataGridWidget(boss, _nfont, xpos, ypos-2, 8, 1, 2, 8, Common::Base::Fmt::_16);
   myTops->setTarget(this);
   myTops->setEditable(false);
 
   // Bottom registers
-  xpos = 18;  ypos += myLineHeight + 4;
-  new StaticTextWidget(boss, _font, xpos, ypos, lwidth,
-        myFontHeight, "Bottom Registers: ", kTextAlignLeft);
+  xpos = 2 + _font.getMaxCharWidth() * 2; ypos += myLineHeight + 4;
+  new StaticTextWidget(boss, _font, xpos, ypos, "Bottom registers ");
   xpos += lwidth;
 
-  myBottoms = new DataGridWidget(boss, _nfont, xpos, ypos-2, 8, 1, 2, 8, Common::Base::F_16);
+  myBottoms = new DataGridWidget(boss, _nfont, xpos, ypos-2, 8, 1, 2, 8, Common::Base::Fmt::_16);
   myBottoms->setTarget(this);
   myBottoms->setEditable(false);
 
   // Counter registers
-  xpos = 18;  ypos += myLineHeight + 4;
-  new StaticTextWidget(boss, _font, xpos, ypos, lwidth,
-        myFontHeight, "Counter Registers: ", kTextAlignLeft);
+  xpos = 2 + _font.getMaxCharWidth() * 2; ypos += myLineHeight + 4;
+  new StaticTextWidget(boss, _font, xpos, ypos, "Counter registers ");
   xpos += lwidth;
 
-  myCounters = new DataGridWidget(boss, _nfont, xpos, ypos-2, 8, 1, 4, 16, Common::Base::F_16_4);
+  myCounters = new DataGridWidget(boss, _nfont, xpos, ypos-2, 8, 1, 4, 16, Common::Base::Fmt::_16_4);
   myCounters->setTarget(this);
   myCounters->setEditable(false);
 
   // Flag registers
-  xpos = 18;  ypos += myLineHeight + 4;
-  new StaticTextWidget(boss, _font, xpos, ypos, lwidth,
-        myFontHeight, "Flag Registers: ", kTextAlignLeft);
+  xpos = 2 + _font.getMaxCharWidth() * 2; ypos += myLineHeight + 4;
+  new StaticTextWidget(boss, _font, xpos, ypos, "Flag registers ");
   xpos += lwidth;
 
-  myFlags = new DataGridWidget(boss, _nfont, xpos, ypos-2, 8, 1, 2, 8, Common::Base::F_16);
+  myFlags = new DataGridWidget(boss, _nfont, xpos, ypos-2, 8, 1, 2, 8, Common::Base::Fmt::_16);
   myFlags->setTarget(this);
   myFlags->setEditable(false);
 
   // Music mode
-  xpos = 10;  ypos += myLineHeight + 12;
-  lwidth = _font.getStringWidth("Music mode (DF5/DF6/DF7): ");
-  new StaticTextWidget(boss, _font, xpos, ypos, lwidth,
-        myFontHeight, "Music mode (DF5/DF6/DF7): ", kTextAlignLeft);
+  xpos = 2; ypos += myLineHeight + V_GAP * 3;
+  lwidth = _font.getStringWidth("Music mode (DF5/DF6/DF7) ");
+  new StaticTextWidget(boss, _font, xpos, ypos, "Music mode (DF5/DF6/DF7) ");
   xpos += lwidth;
 
-  myMusicMode = new DataGridWidget(boss, _nfont, xpos, ypos-2, 3, 1, 2, 8, Common::Base::F_16);
+  myMusicMode = new DataGridWidget(boss, _nfont, xpos, ypos-2, 3, 1, 2, 8, Common::Base::Fmt::_16);
   myMusicMode->setTarget(this);
   myMusicMode->setEditable(false);
 
   // Current random number
-  xpos = 10;  ypos += myLineHeight + 4;
-  new StaticTextWidget(boss, _font, xpos, ypos, lwidth,
-        myFontHeight, "Current random number: ", kTextAlignLeft);
+  xpos = 2; ypos += myLineHeight + V_GAP * 3;
+  new StaticTextWidget(boss, _font, xpos, ypos, "Current random number ");
   xpos += lwidth;
 
-  myRandom = new DataGridWidget(boss, _nfont, xpos, ypos-2, 1, 1, 2, 8, Common::Base::F_16);
+  myRandom = new DataGridWidget(boss, _nfont, xpos, ypos-2, 1, 1, 2, 8, Common::Base::Fmt::_16);
   myRandom->setTarget(this);
   myRandom->setEditable(false);
 }
@@ -153,12 +151,14 @@ void CartridgeDPCWidget::saveOldState()
 
   for(uInt32 i = 0; i < internalRamSize(); ++i)
     myOldState.internalram.push_back(myCart.myDisplayImage[i]);
+
+  myOldState.bank = myCart.getBank();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void CartridgeDPCWidget::loadConfig()
 {
-  myBank->setSelectedIndex(myCart.myCurrentBank);
+  myBank->setSelectedIndex(myCart.getBank(), myCart.getBank() != myOldState.bank);
 
   // Get registers, using change tracking
   IntArray alist;
@@ -172,7 +172,7 @@ void CartridgeDPCWidget::loadConfig()
     changed.push_back(myCart.myTops[i] != myOldState.tops[i]);
   }
   myTops->setList(alist, vlist, changed);
-  
+
   alist.clear();  vlist.clear();  changed.clear();
   for(int i = 0; i < 8; ++i)
   {
@@ -217,9 +217,9 @@ void CartridgeDPCWidget::handleCommand(CommandSender* sender,
 {
   if(cmd == kBankChanged)
   {
-    myCart.unlockBank();
+    myCart.unlockHotspots();
     myCart.bank(myBank->getSelected());
-    myCart.lockBank();
+    myCart.lockHotspots();
     invalidate();
   }
 }
@@ -229,15 +229,14 @@ string CartridgeDPCWidget::bankState()
 {
   ostringstream& buf = buffer();
 
-  static const char* spot[] = { "$FF8", "$FF9" };
-  buf << "Bank = " << std::dec << myCart.myCurrentBank
-      << ", hotspot = " << spot[myCart.myCurrentBank];
+  buf << "Bank #" << std::dec << myCart.getBank()
+      << " (hotspot $" << Common::Base::HEX4 << (0xFFF8 + myCart.getBank()) << ")";
 
   return buf.str();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-uInt32 CartridgeDPCWidget::internalRamSize() 
+uInt32 CartridgeDPCWidget::internalRamSize()
 {
   return 2*1024;
 }
@@ -249,13 +248,13 @@ uInt32 CartridgeDPCWidget::internalRamRPort(int start)
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-string CartridgeDPCWidget::internalRamDescription() 
+string CartridgeDPCWidget::internalRamDescription()
 {
   ostringstream desc;
-  desc << "$0000 - $07FF - 2K display data\n"
-       << "                indirectly accessible to 6507\n"
-       << "                via DPC+'s Data Fetcher registers\n";
-  
+  desc << "2K display data @ $0000 - $" << Common::Base::HEX4 << 0x07FF << "\n"
+    << "  indirectly accessible to 6507 via DPC's\n"
+    << "  data fetcher registers\n";
+
   return desc.str();
 }
 
