@@ -1,102 +1,124 @@
 //============================================================================
 //
-//   SSSS    tt          lll  lll       
-//  SS  SS   tt           ll   ll        
-//  SS     tttttt  eeee   ll   ll   aaaa 
+//   SSSS    tt          lll  lll
+//  SS  SS   tt           ll   ll
+//  SS     tttttt  eeee   ll   ll   aaaa
 //   SSSS    tt   ee  ee  ll   ll      aa
 //      SS   tt   eeeeee  ll   ll   aaaaa  --  "An Atari 2600 VCS Emulator"
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2016 by Bradford W. Mott, Stephen Anthony
+// Copyright (c) 1995-2022 by Bradford W. Mott, Stephen Anthony
 // and the Stella Team
 //
 // See the file "License.txt" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
-//
-// $Id: RamWidget.cxx 3304 2016-04-03 00:35:00Z stephena $
 //============================================================================
 
-#include "DataGridWidget.hxx"
+#include "DataGridRamWidget.hxx"
 #include "EditTextWidget.hxx"
 #include "GuiObject.hxx"
 #include "InputTextDialog.hxx"
 #include "OSystem.hxx"
+#include "Debugger.hxx"
 #include "CartDebug.hxx"
+#include "Font.hxx"
+#include "FBSurface.hxx"
 #include "Widget.hxx"
-
 #include "RamWidget.hxx"
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 RamWidget::RamWidget(GuiObject* boss, const GUI::Font& lfont, const GUI::Font& nfont,
                      int x, int y, int w, int h,
-                     uInt32 ramsize, uInt32 numrows, uInt32 pagesize)
+                     uInt32 ramsize, uInt32 numrows, uInt32 pagesize,
+                     const string& helpAnchor)
   : Widget(boss, lfont, x, y, w, h),
     CommandSender(boss),
-    _nfont(nfont),
-    myFontWidth(lfont.getMaxCharWidth()),
-    myFontHeight(lfont.getFontHeight()),
-    myLineHeight(lfont.getLineHeight()),
-    myButtonHeight(myLineHeight + 4),
-    myCurrentRamBank(0),
-    myRamSize(ramsize),
-    myNumRows(numrows),
-    myPageSize(pagesize)
+    _nfont{nfont},
+    myFontWidth{lfont.getMaxCharWidth()},
+    myFontHeight{lfont.getFontHeight()},
+    myLineHeight{lfont.getLineHeight()},
+    myButtonHeight{static_cast<int>(myLineHeight * 1.25)},
+    myRamSize{ramsize},
+    myNumRows{numrows},
+    myPageSize{pagesize}
 {
-  const int bwidth  = lfont.getStringWidth("Compare "),
+  const int bwidth  = lfont.getStringWidth("Compare " + ELLIPSIS),
             bheight = myLineHeight + 2;
+  //const int VGAP = 4;
+  const int VGAP = myFontHeight / 4;
+  StaticTextWidget* s = nullptr;
+  WidgetArray wid;
 
   int ypos = y + myLineHeight;
 
   // Add RAM grid (with scrollbar)
   int xpos = x + _font.getStringWidth("xxxx");
-  myRamGrid = new DataGridWidget(_boss, _nfont, xpos, ypos,
-                                 16, myNumRows, 2, 8, Common::Base::F_16, true);
+  const bool useScrollbar = ramsize / numrows > 16;
+  myRamGrid = new DataGridRamWidget(_boss, *this, _nfont, xpos, ypos,
+                                    16, myNumRows, 2, 8, Common::Base::Fmt::_16, useScrollbar);
+  myRamGrid->setHelpAnchor(helpAnchor, true);
   myRamGrid->setTarget(this);
-  myRamGrid->setID(kRamHexID);
+  myRamGrid->setID(kRamGridID);
   addFocusWidget(myRamGrid);
 
   // Create actions buttons to the left of the RAM grid
-  int bx = xpos + myRamGrid->getWidth() + 4;
+  const int bx = xpos + myRamGrid->getWidth() + 4;
   int by = ypos;
+
   myUndoButton = new ButtonWidget(boss, lfont, bx, by, bwidth, bheight,
                                   "Undo", kUndoCmd);
+  myUndoButton->setHelpAnchor("M6532Search", true);
+  wid.push_back(myUndoButton);
   myUndoButton->setTarget(this);
 
-  by += bheight + 4;
+  by += bheight + VGAP;
   myRevertButton = new ButtonWidget(boss, lfont, bx, by, bwidth, bheight,
                                     "Revert", kRevertCmd);
+  myRevertButton->setHelpAnchor("M6532Search", true);
+  wid.push_back(myRevertButton);
   myRevertButton->setTarget(this);
 
-  by += 2 * bheight + 2;
+  by += bheight + VGAP * 6;
   mySearchButton = new ButtonWidget(boss, lfont, bx, by, bwidth, bheight,
-                                    "Search", kSearchCmd);
+                                    "Search" + ELLIPSIS, kSearchCmd);
+  mySearchButton->setHelpAnchor("M6532Search", true);
+  mySearchButton->setToolTip("Search and highlight found values.");
+  wid.push_back(mySearchButton);
   mySearchButton->setTarget(this);
 
-  by += bheight + 4;
+  by += bheight + VGAP;
   myCompareButton = new ButtonWidget(boss, lfont, bx, by, bwidth, bheight,
-                                     "Compare", kCmpCmd);
+                                     "Compare" + ELLIPSIS, kCmpCmd);
+  myCompareButton->setHelpAnchor("M6532Search", true);
+  myCompareButton->setToolTip("Compare highlighted values.");
+  wid.push_back(myCompareButton);
   myCompareButton->setTarget(this);
 
-  by += bheight + 4;
+  by += bheight + VGAP;
   myRestartButton = new ButtonWidget(boss, lfont, bx, by, bwidth, bheight,
                                      "Reset", kRestartCmd);
+  myRestartButton->setHelpAnchor("M6532Search", true);
+  myRestartButton->setToolTip("Reset search/compare mode.");
+  wid.push_back(myRestartButton);
   myRestartButton->setTarget(this);
 
+  addToFocusList(wid);
+
   // Labels for RAM grid
   myRamStart =
     new StaticTextWidget(_boss, lfont, xpos - _font.getStringWidth("xxxx"),
                          ypos - myLineHeight,
                          lfont.getStringWidth("xxxx"), myFontHeight,
-                        "00xx", kTextAlignLeft);
+                        "00xx", TextAlign::Left);
 
   for(int col = 0; col < 16; ++col)
   {
     new StaticTextWidget(_boss, lfont, xpos + col*myRamGrid->colWidth() + 8,
                          ypos - myLineHeight,
                          myFontWidth, myFontHeight,
-                         Common::Base::toString(col, Common::Base::F_16_1),
-                         kTextAlignLeft);
+                         Common::Base::toString(col, Common::Base::Fmt::_16_1),
+                         TextAlign::Left);
   }
 
   uInt32 row;
@@ -105,60 +127,73 @@ RamWidget::RamWidget(GuiObject* boss, const GUI::Font& lfont, const GUI::Font& n
     myRamLabels[row] =
       new StaticTextWidget(_boss, _font, xpos - _font.getStringWidth("x "),
                            ypos + row*myLineHeight + 2,
-                           myFontWidth, myFontHeight, "", kTextAlignLeft);
+                           myFontWidth, myFontHeight, "", TextAlign::Left);
   }
 
   // For smaller grids, make sure RAM cell detail fields are below the RESET button
   row = myNumRows < 8 ? 9 : myNumRows + 1;
-  ypos += row * myLineHeight;
+  ypos += (row - 1) * myLineHeight + VGAP * 2;
 
   // We need to define these widgets from right to left since the leftmost
   // one resizes as much as possible
 
   // Add Binary display of selected RAM cell
-  xpos = x + w - 13*myFontWidth - 20;
-  new StaticTextWidget(boss, lfont, xpos, ypos, 4*myFontWidth, myFontHeight,
-                       "Bin:", kTextAlignLeft);
-  myBinValue = new DataGridWidget(boss, nfont, xpos + 4*myFontWidth + 5, ypos-2,
-                                  1, 1, 8, 8, Common::Base::F_2);
+  xpos = x + w - 9.6 * myFontWidth - 9;
+  s = new StaticTextWidget(boss, lfont, xpos, ypos, "%");
+  myBinValue = new DataGridWidget(boss, nfont, s->getRight() + myFontWidth * 0.1, ypos-2,
+                                  1, 1, 8, 8, Common::Base::Fmt::_2);
+  myBinValue->setHelpAnchor(helpAnchor, true);
   myBinValue->setTarget(this);
   myBinValue->setID(kRamBinID);
 
   // Add Decimal display of selected RAM cell
-  xpos -= 8*myFontWidth + 5 + 20;
-  new StaticTextWidget(boss, lfont, xpos, ypos, 4*myFontWidth, myFontHeight,
-                       "Dec:", kTextAlignLeft);
-  myDecValue = new DataGridWidget(boss, nfont, xpos + 4*myFontWidth + 5, ypos-2,
-                                  1, 1, 3, 8, Common::Base::F_10);
+  xpos -= 6.5 * myFontWidth;
+  s = new StaticTextWidget(boss, lfont, xpos, ypos, "#");
+  myDecValue = new DataGridWidget(boss, nfont, s->getRight(), ypos-2,
+                                  1, 1, 3, 8, Common::Base::Fmt::_10);
+  myDecValue->setHelpAnchor(helpAnchor, true);
   myDecValue->setTarget(this);
   myDecValue->setID(kRamDecID);
 
+  // Add Hex display of selected RAM cell
+  xpos -= 4.5 * myFontWidth;
+  myHexValue = new DataGridWidget(boss, nfont, xpos, ypos - 2,
+                                  1, 1, 2, 8, Common::Base::Fmt::_16);
+  myHexValue->setHelpAnchor(helpAnchor, true);
+  myHexValue->setTarget(this);
+  myHexValue->setID(kRamHexID);
+
+  addFocusWidget(myHexValue);
   addFocusWidget(myDecValue);
   addFocusWidget(myBinValue);
 
   // Add Label of selected RAM cell
-  int xpos_r = xpos - 20;
-  xpos = x + 10;
-  new StaticTextWidget(boss, lfont, xpos, ypos, 6*myFontWidth, myFontHeight,
-                       "Label:", kTextAlignLeft);
-  xpos += 6*myFontWidth + 5;
+  const int xpos_r = xpos - myFontWidth * 1.5;
+  xpos = x;
+  s = new StaticTextWidget(boss, lfont, xpos, ypos, "Label");
+  xpos = s->getRight() + myFontWidth / 2;
   myLabel = new EditTextWidget(boss, nfont, xpos, ypos-2, xpos_r-xpos,
                                myLineHeight);
-  myLabel->setEditable(false);
+  myLabel->setEditable(false, true);
 
   // Inputbox which will pop up when searching RAM
-  StringList labels = { "Search: " };
-  myInputBox = make_ptr<InputTextDialog>(boss, lfont, nfont, labels);
+  StringList labels = { "Value" };
+  myInputBox = make_unique<InputTextDialog>(boss, lfont, nfont, labels, " ");
   myInputBox->setTarget(this);
 
   // Start with these buttons disabled
-  myCompareButton->clearFlags(WIDGET_ENABLED);
-  myRestartButton->clearFlags(WIDGET_ENABLED);
+  myCompareButton->clearFlags(Widget::FLAG_ENABLED);
+  myRestartButton->clearFlags(Widget::FLAG_ENABLED);
 
   // Calculate final height
   if(_h == 0)  _h = ypos + myLineHeight - y;
 }
 
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+RamWidget::~RamWidget()
+{
+}
+
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void RamWidget::handleCommand(CommandSender* sender, int cmd, int data, int id)
 {
@@ -173,11 +208,16 @@ void RamWidget::handleCommand(CommandSender* sender, int cmd, int data, int id)
     {
       switch(id)
       {
-        case kRamHexID:
+        case kRamGridID:
           addr  = myRamGrid->getSelectedAddr();
           value = myRamGrid->getSelectedValue();
           break;
 
+        case kRamHexID:
+          addr  = myRamGrid->getSelectedAddr();
+          value = myHexValue->getSelectedValue();
+          break;
+
         case kRamDecID:
           addr  = myRamGrid->getSelectedAddr();
           value = myDecValue->getSelectedValue();
@@ -187,15 +227,19 @@ void RamWidget::handleCommand(CommandSender* sender, int cmd, int data, int id)
           addr  = myRamGrid->getSelectedAddr();
           value = myBinValue->getSelectedValue();
           break;
+
+        default:
+          break;
       }
 
-      uInt8 oldval = getValue(addr);
+      const uInt8 oldval = getValue(addr);
       setValue(addr, value);
 
       myUndoAddress = addr;
       myUndoValue = oldval;
 
       myRamGrid->setValueInternal(addr - myCurrentRamBank*myPageSize, value, true);
+      myHexValue->setValueInternal(0, value, true);
       myDecValue->setValueInternal(0, value, true);
       myBinValue->setValueInternal(0, value, true);
 
@@ -208,10 +252,12 @@ void RamWidget::handleCommand(CommandSender* sender, int cmd, int data, int id)
     {
       addr  = myRamGrid->getSelectedAddr();
       value = myRamGrid->getSelectedValue();
+      const bool changed = myRamGrid->getSelectedChanged();
 
       myLabel->setText(getLabel(addr));
-      myDecValue->setValueInternal(0, value, false);
-      myBinValue->setValueInternal(0, value, false);
+      myHexValue->setValueInternal(0, value, changed);
+      myDecValue->setValueInternal(0, value, changed);
+      myBinValue->setValueInternal(0, value, changed);
       break;
     }
 
@@ -243,7 +289,7 @@ void RamWidget::handleCommand(CommandSender* sender, int cmd, int data, int id)
     {
       const string& result = doSearch(myInputBox->getResult());
       if(result != "")
-        myInputBox->setTitle(result);
+        myInputBox->setMessage(result);
       else
         myInputBox->close();
       break;
@@ -253,17 +299,20 @@ void RamWidget::handleCommand(CommandSender* sender, int cmd, int data, int id)
     {
       const string& result = doCompare(myInputBox->getResult());
       if(result != "")
-        myInputBox->setTitle(result);
+        myInputBox->setMessage(result);
       else
         myInputBox->close();
       break;
     }
 
-    case kSetPositionCmd:
+    case GuiObject::kSetPositionCmd:
       myCurrentRamBank = data;
       showSearchResults();
       fillGrid(false);
       break;
+
+    default:
+      break;
   }
 }
 
@@ -271,6 +320,7 @@ void RamWidget::handleCommand(CommandSender* sender, int cmd, int data, int id)
 void RamWidget::setOpsWidget(DataGridOpsWidget* w)
 {
   myRamGrid->setOpsWidget(w);
+  myHexValue->setOpsWidget(w);
   myBinValue->setOpsWidget(w);
   myDecValue->setOpsWidget(w);
 }
@@ -279,6 +329,13 @@ void RamWidget::setOpsWidget(DataGridOpsWidget* w)
 void RamWidget::loadConfig()
 {
   fillGrid(true);
+
+  const int value = myRamGrid->getSelectedValue();
+  const bool changed = myRamGrid->getSelectedChanged();
+
+  myHexValue->setValueInternal(0, value, changed);
+  myDecValue->setValueInternal(0, value, changed);
+  myBinValue->setValueInternal(0, value, changed);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -288,7 +345,7 @@ void RamWidget::fillGrid(bool updateOld)
   IntArray vlist;
   BoolArray changed;
 
-  uInt32 start = myCurrentRamBank * myPageSize;
+  const uInt32 start = myCurrentRamBank * myPageSize;
   fillList(start, myPageSize, alist, vlist, changed);
 
   if(updateOld)
@@ -303,26 +360,32 @@ void RamWidget::fillGrid(bool updateOld)
   }
 
   // Update RAM labels
-  uInt32 rport = readPort(start), page = rport & 0xf0;
-  char buf[5];
-  std::snprintf(buf, 5, "%04X", rport);
-  buf[2] = buf[3] = 'x';
-  myRamStart->setLabel(buf);
+  const uInt32 rport = readPort(start);
+  int page = rport & 0xf0;
+  string label = Common::Base::toString(rport, Common::Base::Fmt::_16_4);
+
+  label[2] = label[3] = 'x';
+  myRamStart->setLabel(label);
   for(uInt32 row = 0; row < myNumRows; ++row, page += 0x10)
-    myRamLabels[row]->setLabel(Common::Base::toString(page>>4, Common::Base::F_16_1));
+    myRamLabels[row]->setLabel(Common::Base::toString(page>>4, Common::Base::Fmt::_16_1));
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void RamWidget::showInputBox(int cmd)
 {
   // Add inputbox in the middle of the RAM widget
-  uInt32 x = getAbsX() + ((getWidth() - myInputBox->getWidth()) >> 1);
-  uInt32 y = getAbsY() + ((getHeight() - myInputBox->getHeight()) >> 1);
-  myInputBox->show(x, y);
+  const uInt32 x = getAbsX() + ((getWidth() - myInputBox->getWidth()) >> 1);
+  const uInt32 y = getAbsY() + ((getHeight() - myInputBox->getHeight()) >> 1);
+
+  myInputBox->show(x, y, dialog().surface().dstRect());
   myInputBox->setText("");
-  myInputBox->setTitle("");
+  myInputBox->setMessage("");
+  myInputBox->setToolTip(cmd == kSValEntered
+                         ? "Enter search value (leave blank for all)."
+                         : "Enter relative or absolute value\nto compare with searched values.");
   myInputBox->setFocus(0);
   myInputBox->setEmitSignal(cmd);
+  myInputBox->setTitle(cmd == kSValEntered ? "Search" : "Compare");
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -341,7 +404,7 @@ string RamWidget::doSearch(const string& str)
     return "Invalid input +|-";
   }
 
-  int searchVal = instance().debugger().stringToValue(str);
+  const int searchVal = instance().debugger().stringToValue(str);
 
   // Clear the search array of previous items
   mySearchAddr.clear();
@@ -354,7 +417,7 @@ string RamWidget::doSearch(const string& str)
   bool hitfound = false;
   for(uInt32 addr = 0; addr < ram.size(); ++addr)
   {
-    int value = ram[addr];
+    const int value = ram[addr];
     if(comparisonSearch && searchVal != value)
     {
       mySearchState.push_back(false);
@@ -385,25 +448,25 @@ string RamWidget::doSearch(const string& str)
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 string RamWidget::doCompare(const string& str)
 {
-  bool comparitiveSearch = false;
+  bool comparativeSearch = false;
   int searchVal = 0, offset = 0;
 
   if(str.length() == 0)
-    return "Enter an absolute or comparitive value";
+    return "Enter an absolute or comparative value";
 
   // Do some pre-processing on the string
-  string::size_type pos = str.find_first_of("+-", 0);
+  const string::size_type pos = str.find_first_of("+-", 0);
   if(pos > 0 && pos != string::npos)
   {
     // Only accept '+' or '-' at the start of the string
     return "Input must be [+|-]NUM";
   }
 
-  // A comparitive search searches memory for locations that have changed by
+  // A comparative search searches memory for locations that have changed by
   // the specified amount, vs. for exact values
   if(str[0] == '+' || str[0] == '-')
   {
-    comparitiveSearch = true;
+    comparativeSearch = true;
     bool negative = false;
     if(str[0] == '-')
       negative = true;
@@ -427,14 +490,14 @@ string RamWidget::doCompare(const string& str)
 
   for(uInt32 i = 0; i < mySearchAddr.size(); ++i)
   {
-    if(comparitiveSearch)
+    if(comparativeSearch)
     {
       searchVal = mySearchValue[i] + offset;
       if(searchVal < 0 || searchVal > 255)
         continue;
     }
 
-    int addr = mySearchAddr[i];
+    const int addr = mySearchAddr[i];
     if(ram[addr] == searchVal)
     {
       tempAddrList.push_back(addr);
@@ -479,7 +542,7 @@ void RamWidget::showSearchResults()
 {
   // Only update the search results for the bank currently being shown
   BoolArray temp;
-  uInt32 start = myCurrentRamBank * myPageSize;
+  const uInt32 start = myCurrentRamBank * myPageSize;
   if(mySearchState.size() == 0 || start > mySearchState.size())
   {
     for(uInt32 i = 0; i < myPageSize; ++i)
