@@ -8,48 +8,41 @@
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2016 by Bradford W. Mott, Stephen Anthony
+// Copyright (c) 1995-2022 by Bradford W. Mott, Stephen Anthony
 // and the Stella Team
 //
 // See the file "License.txt" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
-//
-// $Id: CartAR.cxx 3316 2016-08-24 23:57:07Z stephena $
 //============================================================================
 
 #include "M6502.hxx"
 #include "System.hxx"
+#include "Settings.hxx"
 #include "CartAR.hxx"
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-CartridgeAR::CartridgeAR(const uInt8* image, uInt32 size,
-                         const Settings& settings)
-  : Cartridge(settings),
-    mySize(std::max(size, 8448u)),
-    myWriteEnabled(false),
-    myPower(true),
-    myPowerRomCycle(0),
-    myDataHoldRegister(0),
-    myNumberOfDistinctAccesses(0),
-    myWritePending(false),
-    myCurrentBank(0)
+CartridgeAR::CartridgeAR(const ByteBuffer& image, size_t size,
+                         const string& md5, const Settings& settings)
+  : Cartridge(settings, md5),
+    mySize{std::max<size_t>(size, LOAD_SIZE)}
 {
   // Create a load image buffer and copy the given image
-  myLoadImages = make_ptr<uInt8[]>(mySize);
-  myNumberOfLoadImages = mySize / 8448;
-  memcpy(myLoadImages.get(), image, size);
+  myLoadImages = make_unique<uInt8[]>(mySize);
+  myNumberOfLoadImages = static_cast<uInt8>(mySize / LOAD_SIZE);
+  std::copy_n(image.get(), size, myLoadImages.get());
 
   // Add header if image doesn't include it
-  if(size < 8448)
-    memcpy(myLoadImages.get()+8192, ourDefaultHeader, 256);
+  if(size < LOAD_SIZE)
+    std::copy_n(ourDefaultHeader.data(), ourDefaultHeader.size(),
+                myLoadImages.get() + myImage.size());
 
-  // We use System::PageAccess.codeAccessBase, but don't allow its use
+  // We use System::PageAccess.romAccessBase, but don't allow its use
   // through a pointer, since the AR scheme doesn't support bankswitching
   // in the normal sense
   //
   // Instead, access will be through the getAccessFlags and setAccessFlags
   // methods below
-  createCodeAccessBase(mySize);
+  createRomAccessArrays(mySize);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -57,16 +50,16 @@ void CartridgeAR::reset()
 {
   // Initialize RAM
 #if 0  // TODO - figure out actual behaviour of the real cart
-  initializeRAM(myImage, 6*1024);
+  initializeRAM(myImage.data(), myImage.size());
+#else
+  myImage.fill(0);
 #endif
-    memset(myImage, 0, 6 * 1024);
 
   // Initialize SC BIOS ROM
   initializeROM();
 
   myWriteEnabled = false;
   myPower = true;
-  myPowerRomCycle = mySystem->cycles();
 
   myDataHoldRegister = 0;
   myNumberOfDistinctAccesses = 0;
@@ -76,22 +69,15 @@ void CartridgeAR::reset()
   bankConfiguration(0);
 }
 
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void CartridgeAR::systemCyclesReset()
-{
-  // Adjust cycle values
-  myPowerRomCycle -= mySystem->cycles();
-}
-
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void CartridgeAR::install(System& system)
 {
   mySystem = &system;
 
   // Map all of the accesses to call peek and poke (we don't yet indicate RAM areas)
-  System::PageAccess access(this, System::PA_READ);
-  for(uInt32 i = 0x1000; i < 0x2000; i += (1 << System::PAGE_SHIFT))
-    mySystem->setPageAccess(i >> System::PAGE_SHIFT, access);
+  const System::PageAccess access(this, System::PageAccessType::READ);
+  for(uInt16 addr = 0x1000; addr < 0x2000; addr += System::PAGE_SIZE)
+    mySystem->setPageAccess(addr, access);
 
   bankConfiguration(0);
 }
@@ -101,14 +87,14 @@ uInt8 CartridgeAR::peek(uInt16 addr)
 {
   // In debugger/bank-locked mode, we ignore all hotspots and in general
   // anything that can change the internal state of the cart
-  if(bankLocked())
+  if(hotspotsLocked())
     return myImage[(addr & 0x07FF) + myImageOffset[(addr & 0x0800) ? 1 : 0]];
 
   // Is the "dummy" SC BIOS hotspot for reading a load being accessed?
-  if(((addr & 0x1FFF) == 0x1850) && (myImageOffset[1] == (3 << 11)))
+  if(((addr & 0x1FFF) == 0x1850) && (myImageOffset[1] == RAM_SIZE))
   {
     // Get load that's being accessed (BIOS places load number at 0x80)
-    uInt8 load = mySystem->peek(0x0080);
+    const uInt8 load = mySystem->peek(0x0080);
 
     // Read the specified load into RAM
     loadIntoRAM(load);
@@ -118,7 +104,7 @@ uInt8 CartridgeAR::peek(uInt16 addr)
 
   // Cancel any pending write if more than 5 distinct accesses have occurred
   // TODO: Modify to handle when the distinct counter wraps around...
-  if(myWritePending && 
+  if(myWritePending &&
       (mySystem->m6502().distinctAccesses() > myNumberOfDistinctAccesses + 5))
   {
     myWritePending = false;
@@ -127,7 +113,7 @@ uInt8 CartridgeAR::peek(uInt16 addr)
   // Is the data hold register being set?
   if(!(addr & 0x0F00) && (!myWriteEnabled || !myWritePending))
   {
-    myDataHoldRegister = addr;
+    myDataHoldRegister = static_cast<uInt8>(addr);  // FIXME - check cast here
     myNumberOfDistinctAccesses = mySystem->m6502().distinctAccesses();
     myWritePending = true;
   }
@@ -139,7 +125,7 @@ uInt8 CartridgeAR::peek(uInt16 addr)
     bankConfiguration(myDataHoldRegister);
   }
   // Handle poke if writing enabled
-  else if(myWriteEnabled && myWritePending && 
+  else if(myWriteEnabled && myWritePending &&
       (mySystem->m6502().distinctAccesses() == (myNumberOfDistinctAccesses + 5)))
   {
     if((addr & 0x0800) == 0)
@@ -147,7 +133,7 @@ uInt8 CartridgeAR::peek(uInt16 addr)
       myImage[(addr & 0x07FF) + myImageOffset[0]] = myDataHoldRegister;
       mySystem->setDirtyPage(addr);
     }
-    else if(myImageOffset[1] != (3 << 11))    // Can't poke to ROM :-)
+    else if(myImageOffset[1] != (3 * BANK_SIZE))    // Can't poke to ROM :-)
     {
       myImage[(addr & 0x07FF) + myImageOffset[1]] = myDataHoldRegister;
       mySystem->setDirtyPage(addr);
@@ -165,7 +151,7 @@ bool CartridgeAR::poke(uInt16 addr, uInt8)
 
   // Cancel any pending write if more than 5 distinct accesses have occurred
   // TODO: Modify to handle when the distinct counter wraps around...
-  if(myWritePending && 
+  if(myWritePending &&
       (mySystem->m6502().distinctAccesses() > myNumberOfDistinctAccesses + 5))
   {
     myWritePending = false;
@@ -174,7 +160,7 @@ bool CartridgeAR::poke(uInt16 addr, uInt8)
   // Is the data hold register being set?
   if(!(addr & 0x0F00) && (!myWriteEnabled || !myWritePending))
   {
-    myDataHoldRegister = addr;
+    myDataHoldRegister = static_cast<uInt8>(addr);  // FIXME - check cast here
     myNumberOfDistinctAccesses = mySystem->m6502().distinctAccesses();
     myWritePending = true;
   }
@@ -186,7 +172,7 @@ bool CartridgeAR::poke(uInt16 addr, uInt8)
     bankConfiguration(myDataHoldRegister);
   }
   // Handle poke if writing enabled
-  else if(myWriteEnabled && myWritePending && 
+  else if(myWriteEnabled && myWritePending &&
       (mySystem->m6502().distinctAccesses() == (myNumberOfDistinctAccesses + 5)))
   {
     if((addr & 0x0800) == 0)
@@ -194,7 +180,7 @@ bool CartridgeAR::poke(uInt16 addr, uInt8)
       myImage[(addr & 0x07FF) + myImageOffset[0]] = myDataHoldRegister;
       modified = true;
     }
-    else if(myImageOffset[1] != (3 << 11))    // Can't poke to ROM :-)
+    else if(myImageOffset[1] != (3 * BANK_SIZE))    // Can't poke to ROM :-)
     {
       myImage[(addr & 0x07FF) + myImageOffset[1]] = myDataHoldRegister;
       modified = true;
@@ -205,19 +191,21 @@ bool CartridgeAR::poke(uInt16 addr, uInt8)
   return modified;
 }
 
+#ifdef DEBUGGER_SUPPORT
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-uInt8 CartridgeAR::getAccessFlags(uInt16 address) const
+Device::AccessFlags CartridgeAR::getAccessFlags(uInt16 address) const
 {
-  return myCodeAccessBase[(address & 0x07FF) +
+  return myRomAccessBase[(address & 0x07FF) +
            myImageOffset[(address & 0x0800) ? 1 : 0]];
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void CartridgeAR::setAccessFlags(uInt16 address, uInt8 flags)
+void CartridgeAR::setAccessFlags(uInt16 address, Device::AccessFlags flags)
 {
-  myCodeAccessBase[(address & 0x07FF) +
+  myRomAccessBase[(address & 0x07FF) +
     myImageOffset[(address & 0x0800) ? 1 : 0]] |= flags;
 }
+#endif
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 bool CartridgeAR::bankConfiguration(uInt8 configuration)
@@ -234,82 +222,27 @@ bool CartridgeAR::bankConfiguration(uInt8 configuration)
   //  101wp     1            ROM
   //  110wp     2            1      as used in Killer Satellites
   //  111wp     1            2      as we use for 2k/4k ROM cloning
-  // 
+  //
   //  w = Write Enable (1 = enabled; accesses to $F000-$F0FF cause writes
   //    to happen.  0 = disabled, and the cart acts like ROM.)
   //  p = ROM Power (0 = enabled, 1 = off.)  Only power the ROM if you're
   //    wanting to access the ROM for multiloads.  Otherwise set to 1.
+  const uInt32 OFFSET_0[8] = {2 * BANK_SIZE, 0 * BANK_SIZE, 2 * BANK_SIZE, 0 * BANK_SIZE,
+                              2 * BANK_SIZE, 1 * BANK_SIZE, 2 * BANK_SIZE, 1 * BANK_SIZE};
+  const uInt32 OFFSET_1[8] = {3 * BANK_SIZE, 3 * BANK_SIZE, 0 * BANK_SIZE, 2 * BANK_SIZE,
+                              3 * BANK_SIZE, 3 * BANK_SIZE, 1 * BANK_SIZE, 2 * BANK_SIZE};
+  const int bankConfig = (configuration & 0b11100) >> 2;
 
-  myCurrentBank = configuration & 0x1F; // remember for the bank() method
+  myCurrentBank = configuration & 0b11111; // remember for the bank() method
 
   // Handle ROM power configuration
-  myPower = !(configuration & 0x01);
+  myPower = !(configuration & 0b00001);
 
-  if(myPower)
-  {
-    myPowerRomCycle = mySystem->cycles();
-  }
+  myWriteEnabled = configuration & 0b00010;
 
-  myWriteEnabled = configuration & 0x02;
+  myImageOffset[0] = OFFSET_0[bankConfig];
+  myImageOffset[1] = OFFSET_1[bankConfig];
 
-  switch((configuration >> 2) & 0x07)
-  {
-    case 0:
-    {
-      myImageOffset[0] = 2 << 11;
-      myImageOffset[1] = 3 << 11;
-      break;
-    }
-
-    case 1:
-    {
-      myImageOffset[0] = 0      ;
-      myImageOffset[1] = 3 << 11;
-      break;
-    }
-
-    case 2:
-    {
-      myImageOffset[0] = 2 << 11;
-      myImageOffset[1] = 0      ;
-      break;
-    }
-
-    case 3:
-    {
-      myImageOffset[0] = 0      ;
-      myImageOffset[1] = 2 << 11;
-      break;
-    }
-
-    case 4:
-    {
-      myImageOffset[0] = 2 << 11;
-      myImageOffset[1] = 3 << 11;
-      break;
-    }
-
-    case 5:
-    {
-      myImageOffset[0] = 1 << 11;
-      myImageOffset[1] = 3 << 11;
-      break;
-    }
-
-    case 6:
-    {
-      myImageOffset[0] = 2 << 11;
-      myImageOffset[1] = 1 << 11;
-      break;
-    }
-
-    case 7:
-    {
-      myImageOffset[0] = 1 << 11;
-      myImageOffset[1] = 2 << 11;
-      break;
-    }
-  }
   return myBankChanged = true;
 }
 
@@ -331,20 +264,20 @@ void CartridgeAR::initializeROM()
   ourDummyROMCode[281] = mySystem->randGenerator().next();
 
   // Initialize ROM with illegal 6502 opcode that causes a real 6502 to jam
-  memset(myImage + (3<<11), 0x02, 2048);
+  std::fill_n(myImage.begin() + (RAM_SIZE), BANK_SIZE, 0x02);
 
   // Copy the "dummy" Supercharger BIOS code into the ROM area
-  memcpy(myImage + (3<<11), ourDummyROMCode, sizeof(ourDummyROMCode));
+  std::copy_n(ourDummyROMCode.data(), ourDummyROMCode.size(), myImage.data() + (RAM_SIZE));
 
   // Finally set 6502 vectors to point to initial load code at 0xF80A of BIOS
-  myImage[(3<<11) + 2044] = 0x0A;
-  myImage[(3<<11) + 2045] = 0xF8;
-  myImage[(3<<11) + 2046] = 0x0A;
-  myImage[(3<<11) + 2047] = 0xF8;
+  myImage[(RAM_SIZE) + BANK_SIZE - 4] = 0x0A;
+  myImage[(RAM_SIZE) + BANK_SIZE - 3] = 0xF8;
+  myImage[(RAM_SIZE) + BANK_SIZE - 2] = 0x0A;
+  myImage[(RAM_SIZE) + BANK_SIZE - 1] = 0xF8;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-uInt8 CartridgeAR::checksum(uInt8* s, uInt16 length)
+uInt8 CartridgeAR::checksum(const uInt8* s, uInt16 length)
 {
   uInt8 sum = 0;
 
@@ -357,44 +290,48 @@ uInt8 CartridgeAR::checksum(uInt8* s, uInt16 length)
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void CartridgeAR::loadIntoRAM(uInt8 load)
 {
+  bool success = true;
   uInt16 image;
 
   // Scan through all of the loads to see if we find the one we're looking for
   for(image = 0; image < myNumberOfLoadImages; ++image)
   {
     // Is this the correct load?
-    if(myLoadImages[(image * 8448) + 8192 + 5] == load)
+    if(myLoadImages[(image * LOAD_SIZE) + myImage.size() + 5] == load)
     {
       // Copy the load's header
-      memcpy(myHeader, myLoadImages.get() + (image * 8448) + 8192, 256);
+      std::copy_n(myLoadImages.get() + (image * LOAD_SIZE) + myImage.size(), myHeader.size(), myHeader.data());
 
-      // Verify the load's header 
-      if(checksum(myHeader, 8) != 0x55)
+      // Verify the load's header
+      if(checksum(myHeader.data(), 8) != 0x55)
       {
         cerr << "WARNING: The Supercharger header checksum is invalid...\n";
+        myMsgCallback("Supercharger load #" + std::to_string(load) + " done with hearder checksum error");
+        success = false;
       }
 
       // Load all of the pages from the load
       bool invalidPageChecksumSeen = false;
       for(uInt32 j = 0; j < myHeader[3]; ++j)
       {
-        uInt32 bank = myHeader[16 + j] & 0x03;
-        uInt32 page = (myHeader[16 + j] >> 2) & 0x07;
-        uInt8* src = myLoadImages.get() + (image * 8448) + (j * 256);
-        uInt8 sum = checksum(src, 256) + myHeader[16 + j] + myHeader[64 + j];
+        uInt32 bank = myHeader[16 + j] & 0b00011;
+        uInt32 page = (myHeader[16 + j] & 0b11100) >> 2;
+        const uInt8* const src = myLoadImages.get() + (image * LOAD_SIZE) + (j * 256);
+        const uInt8 sum = checksum(src, 256) + myHeader[16 + j] + myHeader[64 + j];
 
         if(!invalidPageChecksumSeen && (sum != 0x55))
         {
           cerr << "WARNING: Some Supercharger page checksums are invalid...\n";
+          myMsgCallback("Supercharger load #" + std::to_string(load) + " done with page #"
+                           + std::to_string(j) + " checksum error");
           invalidPageChecksumSeen = true;
         }
 
         // Copy page to Supercharger RAM (don't allow a copy into ROM area)
         if(bank < 3)
-        {
-          memcpy(myImage + (bank * 2048) + (page * 256), src, 256);
-        }
+          std::copy_n(src, 256, myImage.data() + (bank * BANK_SIZE) + (page * 256));
       }
+      success &= !invalidPageChecksumSeen;
 
       // Copy the bank switching byte and starting address into the 2600's
       // RAM for the "dummy" SC BIOS to access it
@@ -403,6 +340,8 @@ void CartridgeAR::loadIntoRAM(uInt8 load)
       mySystem->poke(0x80, myHeader[2]);
 
       myBankChanged = true;
+      if(success)
+        myMsgCallback("Supercharger load #" + std::to_string(load) + " done");
       return;
     }
   }
@@ -413,22 +352,22 @@ void CartridgeAR::loadIntoRAM(uInt8 load)
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-bool CartridgeAR::bank(uInt16 bank)
+bool CartridgeAR::bank(uInt16 bank, uInt16)
 {
-  if(!bankLocked())
-    return bankConfiguration(bank);
+  if(!hotspotsLocked())
+    return bankConfiguration(static_cast<uInt8>(bank));
   else
     return false;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-uInt16 CartridgeAR::getBank() const
+uInt16 CartridgeAR::getBank(uInt16) const
 {
   return myCurrentBank;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-uInt16 CartridgeAR::bankCount() const
+uInt16 CartridgeAR::romBankCount() const
 {
   return 32;
 }
@@ -441,10 +380,10 @@ bool CartridgeAR::patch(uInt16 address, uInt8 value)
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-const uInt8* CartridgeAR::getImage(int& size) const
+const ByteBuffer& CartridgeAR::getImage(size_t& size) const
 {
   size = mySize;
-  return myLoadImages.get();
+  return myLoadImages;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -452,20 +391,18 @@ bool CartridgeAR::save(Serializer& out) const
 {
   try
   {
-    out.putString(name());
-
     // Indicates the offest within the image for the corresponding bank
-    out.putIntArray(myImageOffset, 2);
+    out.putIntArray(myImageOffset.data(), myImageOffset.size());
 
     // The 6K of RAM and 2K of ROM contained in the Supercharger
-    out.putByteArray(myImage, 8192);
+    out.putByteArray(myImage.data(), myImage.size());
 
     // The 256 byte header for the current 8448 byte load
-    out.putByteArray(myHeader, 256);
+    out.putByteArray(myHeader.data(), myHeader.size());
 
-    // All of the 8448 byte loads associated with the game 
+    // All of the 8448 byte loads associated with the game
     // Note that the size of this array is myNumberOfLoadImages * 8448
-    out.putByteArray(myLoadImages.get(), myNumberOfLoadImages * 8448);
+    out.putByteArray(myLoadImages.get(), myNumberOfLoadImages * LOAD_SIZE);
 
     // Indicates how many 8448 loads there are
     out.putByte(myNumberOfLoadImages);
@@ -476,9 +413,6 @@ bool CartridgeAR::save(Serializer& out) const
     // Indicates if the ROM's power is on or off
     out.putBool(myPower);
 
-    // Indicates when the power was last turned on
-    out.putInt(myPowerRomCycle);
-
     // Data hold register used for writing
     out.putByte(myDataHoldRegister);
 
@@ -502,21 +436,18 @@ bool CartridgeAR::load(Serializer& in)
 {
   try
   {
-    if(in.getString() != name())
-      return false;
-
     // Indicates the offest within the image for the corresponding bank
-    in.getIntArray(myImageOffset, 2);
+    in.getIntArray(myImageOffset.data(), myImageOffset.size());
 
     // The 6K of RAM and 2K of ROM contained in the Supercharger
-    in.getByteArray(myImage, 8192);
+    in.getByteArray(myImage.data(), myImage.size());
 
     // The 256 byte header for the current 8448 byte load
-    in.getByteArray(myHeader, 256);
+    in.getByteArray(myHeader.data(), myHeader.size());
 
-    // All of the 8448 byte loads associated with the game 
+    // All of the 8448 byte loads associated with the game
     // Note that the size of this array is myNumberOfLoadImages * 8448
-    in.getByteArray(myLoadImages.get(), myNumberOfLoadImages * 8448);
+    in.getByteArray(myLoadImages.get(), myNumberOfLoadImages * LOAD_SIZE);
 
     // Indicates how many 8448 loads there are
     myNumberOfLoadImages = in.getByte();
@@ -527,9 +458,6 @@ bool CartridgeAR::load(Serializer& in)
     // Indicates if the ROM's power is on or off
     myPower = in.getBool();
 
-    // Indicates when the power was last turned on
-    myPowerRomCycle = in.getInt();
-
     // Data hold register used for writing
     myDataHoldRegister = in.getByte();
 
@@ -549,7 +477,7 @@ bool CartridgeAR::load(Serializer& in)
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-uInt8 CartridgeAR::ourDummyROMCode[] = {
+std::array<uInt8, 294> CartridgeAR::ourDummyROMCode = {
   0xa5, 0xfa, 0x85, 0x80, 0x4c, 0x18, 0xf8, 0xff,
   0xff, 0xff, 0x78, 0xd8, 0xa0, 0x00, 0xa2, 0x00,
   0x94, 0x00, 0xe8, 0xd0, 0xfb, 0x4c, 0x50, 0xf8,
@@ -590,7 +518,7 @@ uInt8 CartridgeAR::ourDummyROMCode[] = {
 };
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-const uInt8 CartridgeAR::ourDefaultHeader[256] = {
+const std::array<uInt8, 256> CartridgeAR::ourDefaultHeader = {
   0xac, 0xfa, 0x0f, 0x18, 0x62, 0x00, 0x24, 0x02,
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
   0x00, 0x04, 0x08, 0x0c, 0x10, 0x14, 0x18, 0x1c,
