@@ -1,20 +1,18 @@
 //============================================================================
 //
-//   SSSS    tt          lll  lll       
-//  SS  SS   tt           ll   ll        
-//  SS     tttttt  eeee   ll   ll   aaaa 
+//   SSSS    tt          lll  lll
+//  SS  SS   tt           ll   ll
+//  SS     tttttt  eeee   ll   ll   aaaa
 //   SSSS    tt   ee  ee  ll   ll      aa
 //      SS   tt   eeeeee  ll   ll   aaaaa  --  "An Atari 2600 VCS Emulator"
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2016 by Bradford W. Mott, Stephen Anthony
+// Copyright (c) 1995-2022 by Bradford W. Mott, Stephen Anthony
 // and the Stella Team
 //
 // See the file "License.txt" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
-//
-// $Id: CartCM.cxx 3316 2016-08-24 23:57:07Z stephena $
 //============================================================================
 
 #include "CompuMate.hxx"
@@ -23,26 +21,27 @@
 #include "CartCM.hxx"
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-CartridgeCM::CartridgeCM(const uInt8* image, uInt32 size, const Settings& settings)
-  : Cartridge(settings),
-    mySWCHA(0xFF),   // portA is all 1's
-    myCurrentBank(0)
+CartridgeCM::CartridgeCM(const ByteBuffer& image, size_t size,
+                         const string& md5, const Settings& settings)
+  : Cartridge(settings, md5),
+    myImage{make_unique<uInt8[]>(16_KB)}
 {
   // Copy the ROM image into my buffer
-  memcpy(myImage, image, std::min(16384u, size));
-  createCodeAccessBase(16384);
-
-  // On powerup, the last bank of ROM is enabled and RAM is disabled
-  myStartBank = mySWCHA & 0x3;
+  std::copy_n(image.get(), std::min(16_KB, size), myImage.get());
+  createRomAccessArrays(16_KB);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void CartridgeCM::reset()
 {
-  initializeRAM(myRAM, 2048);
+  initializeRAM(myRAM.data(), myRAM.size());
+
+  // On powerup, the last bank of ROM is enabled and RAM is disabled
+  mySWCHA = 0xFF;
+  initializeStartBank(mySWCHA & 0x3);
 
   // Upon reset we switch to the startup bank
-  bank(myStartBank);
+  bank(startBank());
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -52,16 +51,16 @@ void CartridgeCM::install(System& system)
 
   // Mirror all access in RIOT; by doing so we're taking responsibility
   // for that address space in peek and poke below.
-  mySystem->m6532().install(system, *this);
+  mySystem->m6532().installDelegate(system, *this);
 
   // Install pages for the startup bank
-  bank(myStartBank);
+  bank(startBank());
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 uInt8 CartridgeCM::peek(uInt16 address)
 {
-  // NOTE: This does not handle accessing cart ROM/RAM, however, this function
+  // NOTE: This does not handle accessing cart ROM/RAM, however, this method
   // should never be called for ROM/RAM because of the way page accessing
   // has been setup (it will only ever be called for RIOT reads)
   return mySystem->m6532().peek(address);
@@ -79,9 +78,14 @@ bool CartridgeCM::poke(uInt16 address, uInt8 value)
     {
       mySWCHA = value;
       bank(mySWCHA & 0x3);
-      uInt8& column = myCompuMate->myColumn;
-      if(value & 0x20) column = 0;
-      if(value & 0x40) column = (column + 1) % 10;
+      if(myCompuMate)
+      {
+        uInt8& column = myCompuMate->column();
+        if(value & 0x20)
+          column = 0;
+        if(value & 0x40)
+          column = (column + 1) % 10;
+      }
     }
     mySystem->m6532().poke(address, value);
   }
@@ -91,17 +95,16 @@ bool CartridgeCM::poke(uInt16 address, uInt8 value)
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 uInt8 CartridgeCM::column() const
 {
-  return myCompuMate->myColumn;
+  return myCompuMate->column();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-bool CartridgeCM::bank(uInt16 bank)
+bool CartridgeCM::bank(uInt16 bank, uInt16)
 {
-  if(bankLocked()) return false;
+  if(hotspotsLocked()) return false;
 
   // Remember what bank we're in
-  myCurrentBank = bank;
-  uInt16 offset = myCurrentBank << 12;
+  myBankOffset = bank << 12;
 
   // Although this scheme contains four 4K ROM banks and one 2K RAM bank,
   // it's easier to think of things in terms of 2K slices, as follows:
@@ -111,53 +114,57 @@ bool CartridgeCM::bank(uInt16 bank)
   // The upper 2K of cart address space can point to either the 2K of RAM or
   // the upper 2K of the current ROM bank
 
-  System::PageAccess access(this, System::PA_READ);
+  System::PageAccess access(this, System::PageAccessType::READ);
 
   // Lower 2K (always ROM)
-  for(uInt32 address = 0x1000; address < 0x1800;
-      address += (1 << System::PAGE_SHIFT))
+  for(uInt16 addr = 0x1000; addr < 0x1800; addr += System::PAGE_SIZE)
   {
-    access.directPeekBase = &myImage[offset + (address & 0x0FFF)];
-    access.codeAccessBase = &myCodeAccessBase[offset + (address & 0x0FFF)];
-    mySystem->setPageAccess(address >> System::PAGE_SHIFT, access);
+    access.directPeekBase = &myImage[myBankOffset + (addr & 0x0FFF)];
+    access.romAccessBase = &myRomAccessBase[myBankOffset + (addr & 0x0FFF)];
+    access.romPeekCounter = &myRomAccessCounter[myBankOffset + (addr & 0x0FFF)];
+    access.romPokeCounter = &myRomAccessCounter[myBankOffset + (addr & 0x0FFF) + myAccessSize];
+    mySystem->setPageAccess(addr, access);
   }
 
   // Upper 2K (RAM or ROM)
-  for(uInt32 address = 0x1800; address < 0x2000;
-      address += (1 << System::PAGE_SHIFT))
+  for(uInt16 addr = 0x1800; addr < 0x2000; addr += System::PAGE_SIZE)
   {
-    access.type = System::PA_READWRITE;
+    access.type = System::PageAccessType::READWRITE;
 
     if(mySWCHA & 0x10)
     {
-      access.directPeekBase = &myImage[offset + (address & 0x0FFF)];
-      access.codeAccessBase = &myCodeAccessBase[offset + (address & 0x0FFF)];
+      access.directPeekBase = &myImage[myBankOffset + (addr & 0x0FFF)];
+      access.romAccessBase = &myRomAccessBase[myBankOffset + (addr & 0x0FFF)];
+      access.romPeekCounter = &myRomAccessCounter[myBankOffset + (addr & 0x0FFF)];
+      access.romPokeCounter = &myRomAccessCounter[myBankOffset + (addr & 0x0FFF) + myAccessSize];
     }
     else
     {
-      access.directPeekBase = &myRAM[address & 0x7FF];
-      access.codeAccessBase = &myCodeAccessBase[offset + (address & 0x07FF)];
+      access.directPeekBase = &myRAM[addr & 0x7FF];
+      access.romAccessBase = &myRomAccessBase[myBankOffset + (addr & 0x07FF)];
+      access.romPeekCounter = &myRomAccessCounter[myBankOffset + (addr & 0x07FF)];
+      access.romPokeCounter = &myRomAccessCounter[myBankOffset + (addr & 0x07FF) + myAccessSize];
     }
 
     if((mySWCHA & 0x30) == 0x20)
-      access.directPokeBase = &myRAM[address & 0x7FF];
+      access.directPokeBase = &myRAM[addr & 0x7FF];
     else
-      access.directPokeBase = 0;
+      access.directPokeBase = nullptr;
 
-    mySystem->setPageAccess(address >> System::PAGE_SHIFT, access);
+    mySystem->setPageAccess(addr, access);
   }
 
   return myBankChanged = true;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-uInt16 CartridgeCM::getBank() const
+uInt16 CartridgeCM::getBank(uInt16) const
 {
-  return myCurrentBank;
+  return myBankOffset >> 12;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-uInt16 CartridgeCM::bankCount() const
+uInt16 CartridgeCM::romBankCount() const
 {
   // We report 4 banks (of ROM), even though RAM can overlap the upper 2K
   // of cart address space at some times
@@ -172,15 +179,15 @@ bool CartridgeCM::patch(uInt16 address, uInt8 value)
   if((mySWCHA & 0x30) == 0x20)
     myRAM[address & 0x7FF] = value;
   else
-    myImage[(myCurrentBank << 12) + address] = value;
+    myImage[myBankOffset + address] = value;
 
   return myBankChanged = true;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-const uInt8* CartridgeCM::getImage(int& size) const
+const ByteBuffer& CartridgeCM::getImage(size_t& size) const
 {
-  size = 16384;
+  size = 16_KB;
   return myImage;
 }
 
@@ -189,11 +196,10 @@ bool CartridgeCM::save(Serializer& out) const
 {
   try
   {
-    out.putString(name());
-    out.putShort(myCurrentBank);
+    out.putShort(myBankOffset);
     out.putByte(mySWCHA);
-    out.putByte(myCompuMate->myColumn);
-    out.putByteArray(myRAM, 2048);
+    out.putByte(myCompuMate->column());
+    out.putByteArray(myRAM.data(), myRAM.size());
   }
   catch(...)
   {
@@ -209,13 +215,10 @@ bool CartridgeCM::load(Serializer& in)
 {
   try
   {
-    if(in.getString() != name())
-      return false;
-
-    myCurrentBank = in.getShort();
+    myBankOffset = in.getShort();
     mySWCHA = in.getByte();
-    myCompuMate->myColumn = in.getByte();
-    in.getByteArray(myRAM, 2048);
+    myCompuMate->column() = in.getByte();
+    in.getByteArray(myRAM.data(), myRAM.size());
   }
   catch(...)
   {
@@ -224,7 +227,7 @@ bool CartridgeCM::load(Serializer& in)
   }
 
   // Remember what bank we were in
-  bank(myCurrentBank);
+  bank(myBankOffset >> 12);
 
   return true;
 }
