@@ -1,88 +1,75 @@
 //============================================================================
 //
-//   SSSS    tt          lll  lll       
-//  SS  SS   tt           ll   ll        
-//  SS     tttttt  eeee   ll   ll   aaaa 
+//   SSSS    tt          lll  lll
+//  SS  SS   tt           ll   ll
+//  SS     tttttt  eeee   ll   ll   aaaa
 //   SSSS    tt   ee  ee  ll   ll      aa
 //      SS   tt   eeeeee  ll   ll   aaaaa  --  "An Atari 2600 VCS Emulator"
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2016 by Bradford W. Mott, Stephen Anthony
+// Copyright (c) 1995-2022 by Bradford W. Mott, Stephen Anthony
 // and the Stella Team
 //
 // See the file "License.txt" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
-//
-// $Id: CartDPC.cxx 3316 2016-08-24 23:57:07Z stephena $
 //============================================================================
 
+#include "Settings.hxx"
 #include "System.hxx"
+#include "AudioSettings.hxx"
 #include "CartDPC.hxx"
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-CartridgeDPC::CartridgeDPC(const uInt8* image, uInt32 size,
-                           const Settings& settings)
-  : Cartridge(settings),
-    mySize(size),
-    mySystemCycles(0),
-    myFractionalClocks(0.0),
-    myCurrentBank(0)
+CartridgeDPC::CartridgeDPC(const ByteBuffer& image, size_t size,
+                           const string& md5, const Settings& settings,
+                           size_t bsSize)
+  : CartridgeF8(image, size, md5, settings, bsSize)
 {
-  // Make a copy of the entire image
-  memcpy(myImage, image, std::min(size, 8192u + 2048u + 256u));
-  createCodeAccessBase(8192);
-
-  // Pointer to the program ROM (8K @ 0 byte offset)
-  myProgramImage = myImage;
-
-  // Pointer to the display ROM (2K @ 8K offset)
-  myDisplayImage = myProgramImage + 8192;
-
-  // Initialize the DPC data fetcher registers
-  for(int i = 0; i < 8; ++i)
-    myTops[i] = myBottoms[i] = myCounters[i] = myFlags[i] = 0;
-
-  // None of the data fetchers are in music mode
-  myMusicMode[0] = myMusicMode[1] = myMusicMode[2] = false;
-
-  // Initialize the DPC's random number generator register (must be non-zero)
-  myRandomNumber = 1;
-
-  // Remember startup bank
-  myStartBank = 1;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void CartridgeDPC::reset()
 {
-  // Update cycles to the current system cycles
-  mySystemCycles = mySystem->cycles();
-  myFractionalClocks = 0.0;
+  CartridgeEnhanced::reset();
 
-  // Upon reset we switch to the startup bank
-  bank(myStartBank);
+  myAudioCycles = 0;
+  myFractionalClocks = 0.0;
+  myDpcPitch = mySettings.getInt(AudioSettings::SETTING_DPC_PITCH);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void CartridgeDPC::systemCyclesReset()
+void CartridgeDPC::consoleChanged(ConsoleTiming timing)
 {
-  // Adjust the cycle counter so that it reflects the new value
-  mySystemCycles -= mySystem->cycles();
+  constexpr double NTSC  = 1193191.66666667;  // NTSC  6507 clock rate
+  constexpr double PAL   = 1182298.0;         // PAL   6507 clock rate
+  constexpr double SECAM = 1187500.0;         // SECAM 6507 clock rate
+
+  switch(timing)
+  {
+    case ConsoleTiming::ntsc:   myClockRate = NTSC;   break;
+    case ConsoleTiming::pal:    myClockRate = PAL;    break;
+    case ConsoleTiming::secam:  myClockRate = SECAM;  break;
+    default:  break;  // satisfy compiler
+  }
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void CartridgeDPC::install(System& system)
 {
-  mySystem = &system;
+  CartridgeEnhanced::install(system);
+
+  myRomOffset = 0x80;
+
+  // Pointer to the display ROM (2K @ 8K offset)
+  myDisplayImage = myImage.get() + 8_KB;
+
+  createRomAccessArrays(8_KB);
 
   // Set the page accessing method for the DPC reading & writing pages
-  System::PageAccess access(this, System::PA_READWRITE);
-  for(uInt32 j = 0x1000; j < 0x1080; j += (1 << System::PAGE_SHIFT))
-    mySystem->setPageAccess(j >> System::PAGE_SHIFT, access);
-
-  // Install pages for the startup bank
-  bank(myStartBank);
+  const System::PageAccess access(this, System::PageAccessType::READWRITE);
+  for(uInt16 addr = 0x1000; addr < 0x1080; addr += System::PAGE_SIZE)
+    mySystem->setPageAccess(addr, access);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -90,16 +77,16 @@ inline void CartridgeDPC::clockRandomNumberGenerator()
 {
   // Table for computing the input bit of the random number generator's
   // shift register (it's the NOT of the EOR of four bits)
-  static constexpr uInt8 f[16] = {
+  static constexpr std::array<uInt8, 16> f = {
     1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1
   };
 
   // Using bits 7, 5, 4, & 3 of the shift register compute the input
   // bit for the shift register
-  uInt8 bit = f[((myRandomNumber >> 3) & 0x07) | 
+  const uInt8 bit = f[((myRandomNumber >> 3) & 0x07) |
       ((myRandomNumber & 0x80) ? 0x08 : 0x00)];
 
-  // Update the shift register 
+  // Update the shift register
   myRandomNumber = (myRandomNumber << 1) | bit;
 }
 
@@ -107,18 +94,16 @@ inline void CartridgeDPC::clockRandomNumberGenerator()
 inline void CartridgeDPC::updateMusicModeDataFetchers()
 {
   // Calculate the number of cycles since the last update
-  Int32 cycles = mySystem->cycles() - mySystemCycles;
-  mySystemCycles = mySystem->cycles();
+  const uInt32 cycles = static_cast<uInt32>(mySystem->cycles() - myAudioCycles);
+  myAudioCycles = mySystem->cycles();
 
   // Calculate the number of DPC OSC clocks since the last update
-  double clocks = ((20000.0 * cycles) / 1193191.66666667) + myFractionalClocks;
-  Int32 wholeClocks = Int32(clocks);
-  myFractionalClocks = clocks - double(wholeClocks);
+  const double clocks = ((myDpcPitch * cycles) / myClockRate) + myFractionalClocks;
+  const uInt32 wholeClocks = static_cast<uInt32>(clocks);
+  myFractionalClocks = clocks - static_cast<double>(wholeClocks);
 
-  if(wholeClocks <= 0)
-  {
+  if(wholeClocks == 0)
     return;
-  }
 
   // Let's update counters and flags of the music mode data fetchers
   for(int x = 5; x <= 7; ++x)
@@ -126,33 +111,25 @@ inline void CartridgeDPC::updateMusicModeDataFetchers()
     // Update only if the data fetcher is in music mode
     if(myMusicMode[x - 5])
     {
-      Int32 top = myTops[x] + 1;
-      Int32 newLow = Int32(myCounters[x] & 0x00ff);
+      const Int32 top = myTops[x] + 1;
+      Int32 newLow = static_cast<Int32>(myCounters[x] & 0x00ff);
 
       if(myTops[x] != 0)
       {
         newLow -= (wholeClocks % top);
         if(newLow < 0)
-        {
           newLow += top;
-        }
       }
       else
-      {
         newLow = 0;
-      }
 
       // Update flag register for this data fetcher
       if(newLow <= myBottoms[x])
-      {
         myFlags[x] = 0x00;
-      }
       else if(newLow <= myTops[x])
-      {
         myFlags[x] = 0xff;
-      }
 
-      myCounters[x] = (myCounters[x] & 0x0700) | uInt16(newLow);
+      myCounters[x] = (myCounters[x] & 0x0700) | static_cast<uInt16>(newLow);
     }
   }
 }
@@ -160,15 +137,17 @@ inline void CartridgeDPC::updateMusicModeDataFetchers()
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 uInt8 CartridgeDPC::peek(uInt16 address)
 {
+  const uInt16 peekAddress = address;
+
   address &= 0x0FFF;
 
   // In debugger/bank-locked mode, we ignore all hotspots and in general
   // anything that can change the internal state of the cart
-  if(bankLocked())
-    return myProgramImage[(myCurrentBank << 12) + address];
+  if(hotspotsLocked())
+    return myImage[myCurrentSegOffset[0] + address];
 
   // Clock the random number generator.  This should be done for every
-  // cartridge access, however, we're only doing it for the DPC and 
+  // cartridge access, however, we're only doing it for the DPC and
   // hot-spot accesses to save time.
   clockRandomNumberGenerator();
 
@@ -177,8 +156,8 @@ uInt8 CartridgeDPC::peek(uInt16 address)
     uInt8 result = 0;
 
     // Get the index of the data fetcher that's being accessed
-    uInt32 index = address & 0x07;
-    uInt32 function = (address >> 3) & 0x07;
+    const uInt32 index = address & 0x07;
+    const uInt32 function = (address >> 3) & 0x07;
 
     // Update flag register for selected data fetcher
     if((myCounters[index] & 0x00ff) == myTops[index])
@@ -202,7 +181,7 @@ uInt8 CartridgeDPC::peek(uInt16 address)
         // No, it's a music read
         else
         {
-          static constexpr uInt8 musicAmplitudes[8] = {
+          static constexpr std::array<uInt8, 8> musicAmplitudes = {
               0x00, 0x04, 0x05, 0x09, 0x06, 0x0a, 0x0b, 0x0f
           };
 
@@ -256,7 +235,7 @@ uInt8 CartridgeDPC::peek(uInt16 address)
     }
 
     // Clock the selected data fetcher's counter if needed
-    if((index < 5) || ((index >= 5) && (!myMusicMode[index - 5])))
+    if(index < 5 || !myMusicMode[index - 5])
     {
       myCounters[index] = (myCounters[index] - 1) & 0x07ff;
     }
@@ -264,42 +243,26 @@ uInt8 CartridgeDPC::peek(uInt16 address)
     return result;
   }
   else
-  {
-    // Switch banks if necessary
-    switch(address)
-    {
-      case 0x0FF8:
-        // Set the current bank to the lower 4k bank
-        bank(0);
-        break;
-
-      case 0x0FF9:
-        // Set the current bank to the upper 4k bank
-        bank(1);
-        break;
-
-      default:
-        break;
-    }
-    return myProgramImage[(myCurrentBank << 12) + address];
-  }
+    return CartridgeEnhanced::peek(peekAddress);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 bool CartridgeDPC::poke(uInt16 address, uInt8 value)
 {
+  const uInt16 pokeAddress = address;
+
   address &= 0x0FFF;
 
   // Clock the random number generator.  This should be done for every
-  // cartridge access, however, we're only doing it for the DPC and 
+  // cartridge access, however, we're only doing it for the DPC and
   // hot-spot accesses to save time.
   clockRandomNumberGenerator();
 
   if((address >= 0x0040) && (address < 0x0080))
   {
     // Get the index of the data fetcher that's being accessed
-    uInt32 index = address & 0x07;    
-    uInt32 function = (address >> 3) & 0x07;
+    const uInt32 index = address & 0x07;
+    const uInt32 function = (address >> 3) & 0x07;
 
     switch(function)
     {
@@ -326,14 +289,14 @@ bool CartridgeDPC::poke(uInt16 address, uInt8 value)
           // Data fetcher is in music mode so its low counter value
           // should be loaded from the top register not the poked value
           myCounters[index] = (myCounters[index] & 0x0700) |
-              uInt16(myTops[index]);
+            static_cast<uInt16>(myTops[index]);
         }
         else
         {
           // Data fetcher is either not a music mode data fetcher or it
           // isn't in music mode so it's low counter value should be loaded
           // with the poked value
-          myCounters[index] = (myCounters[index] & 0x0700) | uInt16(value);
+          myCounters[index] = (myCounters[index] & 0x0700) | static_cast<uInt16>(value);
         }
         break;
       }
@@ -341,7 +304,7 @@ bool CartridgeDPC::poke(uInt16 address, uInt8 value)
       // DFx counter high
       case 0x03:
       {
-        myCounters[index] = ((uInt16(value) & 0x07) << 8) |
+        myCounters[index] = ((static_cast<uInt16>(value) & 0x07) << 8) |
             (myCounters[index] & 0x00ff);
 
         // Execute special code for music mode data fetchers
@@ -364,128 +327,56 @@ bool CartridgeDPC::poke(uInt16 address, uInt8 value)
       }
 
       default:
-      {
         break;
-      }
     }
   }
   else
-  {
-    // Switch banks if necessary
-    switch(address)
-    {
-      case 0x0FF8:
-        // Set the current bank to the lower 4k bank
-        bank(0);
-        break;
+    CartridgeEnhanced::poke(pokeAddress, value);
 
-      case 0x0FF9:
-        // Set the current bank to the upper 4k bank
-        bank(1);
-        break;
-
-      default:
-        break;
-    }
-  }
   return false;
 }
 
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-bool CartridgeDPC::bank(uInt16 bank)
-{
-  if(bankLocked()) return false;
-
-  // Remember what bank we're in
-  myCurrentBank = bank;
-  uInt16 offset = myCurrentBank << 12;
-
-  System::PageAccess access(this, System::PA_READ);
-
-  // Set the page accessing methods for the hot spots
-  for(uInt32 i = (0x1FF8 & ~System::PAGE_MASK); i < 0x2000;
-      i += (1 << System::PAGE_SHIFT))
-  {
-    access.codeAccessBase = &myCodeAccessBase[offset + (i & 0x0FFF)];
-    mySystem->setPageAccess(i >> System::PAGE_SHIFT, access);
-  }
-
-  // Setup the page access methods for the current bank
-  for(uInt32 address = 0x1080; address < (0x1FF8U & ~System::PAGE_MASK);
-      address += (1 << System::PAGE_SHIFT))
-  {
-    access.directPeekBase = &myProgramImage[offset + (address & 0x0FFF)];
-    access.codeAccessBase = &myCodeAccessBase[offset + (address & 0x0FFF)];
-    mySystem->setPageAccess(address >> System::PAGE_SHIFT, access);
-  }
-  return myBankChanged = true;
-}
-
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-uInt16 CartridgeDPC::getBank() const
-{
-  return myCurrentBank;
-}
-
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-uInt16 CartridgeDPC::bankCount() const
-{
-  return 2;
-}
-
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 bool CartridgeDPC::patch(uInt16 address, uInt8 value)
 {
-  address &= 0x0FFF;
-
   // For now, we ignore attempts to patch the DPC address space
-  if(address >= 0x0080)
+  if((address & ADDR_MASK) >= ROM_OFFSET + myRomOffset)
   {
-    myProgramImage[(myCurrentBank << 12) + (address & 0x0FFF)] = value;
-    return myBankChanged = true;
+    return CartridgeEnhanced::patch(address, value);
   }
   else
     return false;
 }
 
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-const uInt8* CartridgeDPC::getImage(int& size) const
-{
-  size = mySize;
-  return myImage;
-}
-
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 bool CartridgeDPC::save(Serializer& out) const
 {
+  if(!CartridgeEnhanced::save(out))
+    return false;
+
   try
   {
-    out.putString(name());
-
-    // Indicates which bank is currently active
-    out.putShort(myCurrentBank);
-
     // The top registers for the data fetchers
-    out.putByteArray(myTops, 8);
+    out.putByteArray(myTops.data(), myTops.size());
 
     // The bottom registers for the data fetchers
-    out.putByteArray(myBottoms, 8);
+    out.putByteArray(myBottoms.data(), myBottoms.size());
 
     // The counter registers for the data fetchers
-    out.putShortArray(myCounters, 8);
+    out.putShortArray(myCounters.data(), myCounters.size());
 
     // The flag registers for the data fetchers
-    out.putByteArray(myFlags, 8);
+    out.putByteArray(myFlags.data(), myFlags.size());
 
     // The music mode flags for the data fetchers
-    for(int i = 0; i < 3; ++i)
-      out.putBool(myMusicMode[i]);
+    for(const auto& mode: myMusicMode)
+      out.putBool(mode);
 
     // The random number generator register
     out.putByte(myRandomNumber);
 
-    out.putInt(mySystemCycles);
-    out.putInt(uInt32(myFractionalClocks * 100000000.0));
+    out.putLong(myAudioCycles);
+    out.putDouble(myFractionalClocks);
   }
   catch(...)
   {
@@ -499,45 +390,38 @@ bool CartridgeDPC::save(Serializer& out) const
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 bool CartridgeDPC::load(Serializer& in)
 {
+  if(!CartridgeEnhanced::load(in))
+    return false;
+
   try
   {
-    if(in.getString() != name())
-      return false;
-
-    // Indicates which bank is currently active
-    myCurrentBank = in.getShort();
-
     // The top registers for the data fetchers
-    in.getByteArray(myTops, 8);
+    in.getByteArray(myTops.data(), myTops.size());
 
     // The bottom registers for the data fetchers
-    in.getByteArray(myBottoms, 8);
+    in.getByteArray(myBottoms.data(), myBottoms.size());
 
     // The counter registers for the data fetchers
-    in.getShortArray(myCounters, 8);
+    in.getShortArray(myCounters.data(), myCounters.size());
 
     // The flag registers for the data fetchers
-    in.getByteArray(myFlags, 8);
+    in.getByteArray(myFlags.data(), myFlags.size());
 
     // The music mode flags for the data fetchers
-    for(int i = 0; i < 3; ++i)
-      myMusicMode[i] = in.getBool();
+    for(auto& mode: myMusicMode)
+      mode = in.getBool();
 
     // The random number generator register
     myRandomNumber = in.getByte();
 
     // Get system cycles and fractional clocks
-    mySystemCycles = Int32(in.getInt());
-    myFractionalClocks = double(in.getInt()) / 100000000.0;
+    myAudioCycles = in.getLong();
+    myFractionalClocks = in.getDouble();
   }
   catch(...)
   {
     cerr << "ERROR: CartridgeDPC::load" << endl;
     return false;
   }
-
-  // Now, go to the current bank
-  bank(myCurrentBank);
-
   return true;
 }
