@@ -8,106 +8,126 @@
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2016 by Bradford W. Mott, Stephen Anthony
+// Copyright (c) 1995-2022 by Bradford W. Mott, Stephen Anthony
 // and the Stella Team
 //
 // See the file "License.txt" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
-//
-// $Id: CartDPCPlus.cxx 3316 2016-08-24 23:57:07Z stephena $
 //============================================================================
 
 #ifdef DEBUGGER_SUPPORT
   #include "Debugger.hxx"
 #endif
+#include "MD5.hxx"
 #include "System.hxx"
-#include "Thumbulator.hxx"
 #include "CartDPCPlus.hxx"
+#include "TIA.hxx"
+#include "exception/FatalEmulationError.hxx"
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-CartridgeDPCPlus::CartridgeDPCPlus(const uInt8* image, uInt32 size,
-                                   const Settings& settings)
-  : Cartridge(settings),
-    myFastFetch(false),
-    myLDAimmediate(false),
-    myParameterPointer(0),
-    mySystemCycles(0),
-    myFractionalClocks(0.0),
-    myCurrentBank(0)
+CartridgeDPCPlus::CartridgeDPCPlus(const ByteBuffer& image, size_t size,
+                                   const string& md5, const Settings& settings)
+  : CartridgeARM(md5, settings),
+    myImage{make_unique<uInt8[]>(32_KB)},
+    mySize{std::min(size, 32_KB)}
 {
-  // Store image, making sure it's at least 29KB
-  uInt32 minsize = 4096 * 6 + 4096 + 1024 + 255;
-  mySize = std::max(minsize, size);
-  myImage = make_ptr<uInt8[]>(mySize);
-  memcpy(myImage.get(), image, size);
-  createCodeAccessBase(4096 * 6);
+  // Image is always 32K, but in the case of ROM < 32K, the image is
+  // copied to the end of the buffer
+  if(mySize < 32_KB)
+    std::fill_n(myImage.get(), mySize, 0);
+  std::copy_n(image.get(), size, myImage.get() + (32_KB - mySize));
+  createRomAccessArrays(24_KB);
 
-  // Pointer to the program ROM (24K @ 0 byte offset)
-  myProgramImage = myImage.get();
+  // Pointer to the program ROM (24K @ 3K offset; ignore first 3K)
+  myProgramImage = myImage.get() + 3_KB;
 
   // Pointer to the display RAM
-  myDisplayImage = myDPCRAM + 0xC00;
+  myDisplayImage = myDPCRAM.data() + 3_KB;
 
   // Pointer to the Frequency RAM
-  myFrequencyImage = myDisplayImage + 0x1000;
+  myFrequencyImage = myDisplayImage + 4_KB;
 
-  // If the image is larger than 29K, we assume any excess at the
-  // beginning is ARM code, and skip over it
-  if(size > 29 * 1024)
-    myProgramImage += (size - 29 * 1024);
-
-#ifdef THUMB_SUPPORT
   // Create Thumbulator ARM emulator
-  myThumbEmulator = make_ptr<Thumbulator>
-      (reinterpret_cast<uInt16*>(myProgramImage-0xC00),
-       reinterpret_cast<uInt16*>(myDPCRAM),
-       settings.getBool("thumb.trapfatal"));
-#endif
-  setInitialState();
+  bool devSettings = settings.getBool("dev.settings");
+  myThumbEmulator = make_unique<Thumbulator>
+      (reinterpret_cast<uInt16*>(myImage.get()),
+       reinterpret_cast<uInt16*>(myDPCRAM.data()),
+       static_cast<uInt32>(32_KB),
+      0x00000C00,
+      0x00000C08,
+      0x40001FFC,
+       devSettings ? settings.getBool("dev.thumb.trapfatal") : false,
+       devSettings ? static_cast<double>(
+          settings.getFloat("dev.thumb.cyclefactor")) : 1.0,
+       Thumbulator::ConfigureFor::DPCplus,
+       this);
 
-  // DPC+ always starts in bank 5
-  myStartBank = 5;
+  // Currently 4 DPC+ driver versions have been identified:
+  //   17884ec14f9b1d06fe8d617a1fbdcf47  Jitter  Encore Compatible
+  //   5f80b5a5adbe483addc3f6e6f1b472f8  Stable  Encore Compatible
+  //   8dd73b44fd11c488326ce507cbeb19d1  Stable  NOT Encore Compatible
+  //   b328dbdf787400c0f0e2b88b425872a5  Jitter  Encore Compatible
+  //
+  // Jitter/Stable refers to the appearance of the playfield in bB games if
+  // the DFxFRACINC registers are not updated before every drawscreen.
+  //
+  // The default mask for DFxFRACLOW implements the Jitter behavior. This
+  // changes the mask to implement the Stable behavior.
+  myDriverMD5 = MD5::hash(image, 3_KB);
+  if(myDriverMD5 == "5f80b5a5adbe483addc3f6e6f1b472f8" ||
+     myDriverMD5 == "8dd73b44fd11c488326ce507cbeb19d1" )
+    myFractionalLowMask = 0x0F0000;
+
+  this->setInitialState();
+
+  myPlusROM = make_unique<PlusROM>(mySettings, *this);
+
+  // Determine whether we have a PlusROM cart
+  myPlusROM->initialize(myImage, mySize);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void CartridgeDPCPlus::reset()
 {
-  // Update cycles to the current system cycles
-  mySystemCycles = mySystem->cycles();
-  myFractionalClocks = 0.0;
-
   setInitialState();
 
+  // DPC+ always starts in bank 5
+  initializeStartBank(5);
+
   // Upon reset we switch to the startup bank
-  bank(myStartBank);
+  bank(startBank());
+
+  CartridgeARM::reset();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void CartridgeDPCPlus::setInitialState()
 {
   // Reset various ROM and RAM locations
-  memset(myDPCRAM, 0, 8192);
+  myDPCRAM.fill(0);
 
   // Copy initial DPC display data and Frequency table state to Harmony RAM
-  memcpy(myDisplayImage, myProgramImage + 0x6000, 0x1400);
+  std::copy_n(myProgramImage + 24_KB, 5_KB, myDisplayImage);
 
   // Initialize the DPC data fetcher registers
-  for(int i = 0; i < 8; ++i)
-    myTops[i] = myBottoms[i] = myCounters[i] = myFractionalIncrements[i] = 
-    myFractionalCounters[i] = 0;
+  myTops.fill(0);
+  myBottoms.fill(0);
+  myFractionalIncrements.fill(0);
+  myFractionalCounters.fill(0);
+  myCounters.fill(0);
 
   // Set waveforms to first waveform entry
-  myMusicWaveforms[0] = myMusicWaveforms[1] = myMusicWaveforms[2] = 0;
+  myMusicWaveforms.fill(0);
 
   // Initialize the DPC's random number generator register (must be non-zero)
   myRandomNumber = 0x2B435044; // "DPC+"
-}
 
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void CartridgeDPCPlus::systemCyclesReset()
-{
-  // Adjust the cycle counter so that it reflects the new value
-  mySystemCycles -= mySystem->cycles();
+  // Initialize various other parameters
+  myFastFetch = myLDAimmediate = false;
+  myAudioCycles = myARMCycles = 0;
+  myFractionalClocks = 0.0;
+
+  CartridgeARM::setInitialState();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -116,12 +136,12 @@ void CartridgeDPCPlus::install(System& system)
   mySystem = &system;
 
   // Map all of the accesses to call peek and poke
-  System::PageAccess access(this, System::PA_READ);
-  for(uInt32 i = 0x1000; i < 0x1080; i += (1 << System::PAGE_SHIFT))
-    mySystem->setPageAccess(i >> System::PAGE_SHIFT, access);
+  const System::PageAccess access(this, System::PageAccessType::READ);
+  for(uInt16 addr = 0x1000; addr < 0x1080; addr += System::PAGE_SIZE)
+    mySystem->setPageAccess(addr, access);
 
   // Install pages for the startup bank
-  bank(myStartBank);
+  bank(startBank());
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -136,7 +156,7 @@ inline void CartridgeDPCPlus::clockRandomNumberGenerator()
 inline void CartridgeDPCPlus::priorClockRandomNumberGenerator()
 {
   // Update random number generator (32-bit LFSR, reversed)
-  myRandomNumber = ((myRandomNumber & (1u<<31)) ?
+  myRandomNumber = ((myRandomNumber & (1U<<31)) ?
     ((0x10adab1e^myRandomNumber) << 11) | ((0x10adab1e^myRandomNumber) >> 21) :
     (myRandomNumber << 11) | (myRandomNumber >> 21));
 }
@@ -145,31 +165,25 @@ inline void CartridgeDPCPlus::priorClockRandomNumberGenerator()
 inline void CartridgeDPCPlus::updateMusicModeDataFetchers()
 {
   // Calculate the number of cycles since the last update
-  Int32 cycles = mySystem->cycles() - mySystemCycles;
-  mySystemCycles = mySystem->cycles();
+  const uInt32 cycles = static_cast<uInt32>(mySystem->cycles() - myAudioCycles);
+  myAudioCycles = mySystem->cycles();
 
-  // Calculate the number of DPC OSC clocks since the last update
-  double clocks = ((20000.0 * cycles) / 1193191.66666667) + myFractionalClocks;
-  Int32 wholeClocks = Int32(clocks);
-  myFractionalClocks = clocks - double(wholeClocks);
-
-  if(wholeClocks <= 0)
-  {
-    return;
-  }
+  // Calculate the number of DPC+ OSC clocks since the last update
+  const double clocks = ((20000.0 * cycles) / myClockRate) + myFractionalClocks;
+  const uInt32 wholeClocks = static_cast<uInt32>(clocks);
+  myFractionalClocks = clocks - static_cast<double>(wholeClocks);
 
   // Let's update counters and flags of the music mode data fetchers
-  for(int x = 0; x <= 2; ++x)
-  {
-    myMusicCounters[x] += myMusicFrequencies[x];
-  }
+  if(wholeClocks > 0)
+    for(int x = 0; x <= 2; ++x)
+      myMusicCounters[x] += myMusicFrequencies[x] * wholeClocks;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 inline void CartridgeDPCPlus::callFunction(uInt8 value)
 {
   // myParameter
-  uInt16 ROMdata = (myParameter[1] << 8) + myParameter[0];
+  const uInt16 ROMdata = (myParameter[1] << 8) + myParameter[0];
   switch (value)
   {
     case 0: // Parameter Pointer reset
@@ -185,40 +199,47 @@ inline void CartridgeDPCPlus::callFunction(uInt8 value)
         myDisplayImage[myCounters[myParameter[2]]+i] = myParameter[0];
       myParameterPointer = 0;
       break;
-  #ifdef THUMB_SUPPORT
-    case 254:
-    case 255:
       // Call user written ARM code (most likely be C compiled for ARM)
+    case 254: // call with IRQ driven audio, no special handling needed at this
+              // time for Stella as ARM code "runs in zero 6507 cycles".
+    case 255: // call without IRQ driven audio
       try {
-        myThumbEmulator->run();
+        uInt32 cycles = static_cast<uInt32>(mySystem->cycles() - myARMCycles);
+
+        myARMCycles = mySystem->cycles();
+        myThumbEmulator->run(cycles, value == 254);
+        updateCycles(cycles);
       }
       catch(const runtime_error& e) {
         if(!mySystem->autodetectMode())
         {
-      #ifdef DEBUGGER_SUPPORT
-          Debugger::debugger().startWithFatalError(e.what());
-      #else
-          cout << e.what() << endl;
-      #endif
+          FatalEmulationError::raise(e.what());
         }
       }
       break;
-  #endif
-    // reserved
+    default:  // reserved
+      break;
   }
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 uInt8 CartridgeDPCPlus::peek(uInt16 address)
 {
+  // Is this a PlusROM?
+  if(myPlusROM->isValid())
+  {
+    uInt8 value = 0;
+    if(myPlusROM->peekHotspot(address, value))
+      return value;
+  }
+
   address &= 0x0FFF;
 
-  uInt8 peekvalue = myProgramImage[(myCurrentBank << 12) + address];
-  uInt8 flag;
+  const uInt8 peekvalue = myProgramImage[myBankOffset + address];
 
   // In debugger/bank-locked mode, we ignore all hotspots and in general
   // anything that can change the internal state of the cart
-  if(bankLocked())
+  if(hotspotsLocked())
     return peekvalue;
 
   // Check if we're in Fast Fetch mode and the prior byte was an A9 (LDA #value)
@@ -235,12 +256,12 @@ uInt8 CartridgeDPCPlus::peek(uInt16 address)
     uInt8 result = 0;
 
     // Get the index of the data fetcher that's being accessed
-    uInt32 index = address & 0x07;
-    uInt32 function = (address >> 3) & 0x07;
+    const uInt32 index = address & 0x07;
+    const uInt32 function = (address >> 3) & 0x07;
+
+    // Update flag for selected data fetcher
+    const uInt8 flag = (((myTops[index]-(myCounters[index] & 0x00ff)) & 0xFF) > ((myTops[index]-myBottoms[index]) & 0xFF)) ? 0xFF : 0;
 
-    // Update flag for selected data fetcher  
-    flag = (((myTops[index]-(myCounters[index] & 0x00ff)) & 0xFF) > ((myTops[index]-myBottoms[index]) & 0xFF)) ? 0xFF : 0;
-    
     switch(function)
     {
       case 0x00:
@@ -276,17 +297,21 @@ uInt8 CartridgeDPCPlus::peek(uInt16 address)
 
             // using myDisplayImage[] instead of myProgramImage[] because waveforms
             // can be modified during runtime.
-            uInt32 i = myDisplayImage[(myMusicWaveforms[0] << 5) + (myMusicCounters[0] >> 27)] +
-                       myDisplayImage[(myMusicWaveforms[1] << 5) + (myMusicCounters[1] >> 27)] +
-                       myDisplayImage[(myMusicWaveforms[2] << 5) + (myMusicCounters[2] >> 27)];
+            const uInt32 i =
+                myDisplayImage[(myMusicWaveforms[0] << 5) + (myMusicCounters[0] >> 27)] +
+                myDisplayImage[(myMusicWaveforms[1] << 5) + (myMusicCounters[1] >> 27)] +
+                myDisplayImage[(myMusicWaveforms[2] << 5) + (myMusicCounters[2] >> 27)];
 
-            result = uInt8(i);
+            result = static_cast<uInt8>(i);
             break;
           }
 
           case 0x06:  // reserved
           case 0x07:  // reserved
             break;
+
+          default:
+            break;
         }
         break;
       }
@@ -332,6 +357,8 @@ uInt8 CartridgeDPCPlus::peek(uInt16 address)
           case 0x06:  // reserved
           case 0x07:  // reserved
             break;
+          default:
+            break;
         }
         break;
       }
@@ -393,24 +420,30 @@ uInt8 CartridgeDPCPlus::peek(uInt16 address)
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 bool CartridgeDPCPlus::poke(uInt16 address, uInt8 value)
 {
+  // Is this a PlusROM?
+  if(myPlusROM->isValid() && myPlusROM->pokeHotspot(address, value))
+    return true;
+
   address &= 0x0FFF;
 
   if((address >= 0x0028) && (address < 0x0080))
   {
     // Get the index of the data fetcher that's being accessed
-    uInt32 index = address & 0x07;
-    uInt32 function = ((address - 0x28) >> 3) & 0x0f;
+    const uInt32 index = address & 0x07;
+    const uInt32 function = ((address - 0x28) >> 3) & 0x0f;
 
     switch(function)
     {
-      //DFxFRACLOW - fractional data pointer low byte
+      // DFxFRACLOW - fractional data pointer low byte
       case 0x00:
-        myFractionalCounters[index] = (myFractionalCounters[index] & 0x0F00FF) | (uInt16(value) << 8);
+        myFractionalCounters[index] =
+          (myFractionalCounters[index] & myFractionalLowMask) | (static_cast<uInt16>(value) << 8);
         break;
 
       // DFxFRACHI - fractional data pointer high byte
       case 0x01:
-        myFractionalCounters[index] = ((uInt16(value) & 0x0F) << 16) | (myFractionalCounters[index] & 0x00ffff);
+        myFractionalCounters[index] = ((static_cast<uInt16>(value) & 0x0F) << 16) |
+                                       (myFractionalCounters[index] & 0x00ffff);
         break;
 
       //DFxFRACINC - Fractional Increment amount
@@ -460,6 +493,8 @@ bool CartridgeDPCPlus::poke(uInt16 address, uInt8 value)
           case 0x07:  // WAVEFORM2
             myMusicWaveforms[index - 5] =  value & 0x7f;
             break;
+          default:
+            break;
         }
         break;
 
@@ -474,7 +509,7 @@ bool CartridgeDPCPlus::poke(uInt16 address, uInt8 value)
       // DFxHI - data pointer high byte
       case 0x08:
       {
-        myCounters[index] = ((uInt16(value) & 0x0F) << 8) | (myCounters[index] & 0x00ff);
+        myCounters[index] = ((static_cast<uInt16>(value) & 0x0F) << 8) | (myCounters[index] & 0x00ff);
         break;
       }
 
@@ -532,9 +567,7 @@ bool CartridgeDPCPlus::poke(uInt16 address, uInt8 value)
       }
 
       default:
-      {
         break;
-      }
     }
   }
   else
@@ -580,35 +613,35 @@ bool CartridgeDPCPlus::poke(uInt16 address, uInt8 value)
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-bool CartridgeDPCPlus::bank(uInt16 bank)
+bool CartridgeDPCPlus::bank(uInt16 bank, uInt16)
 {
-  if(bankLocked()) return false;
+  if(hotspotsLocked()) return false;
 
   // Remember what bank we're in
-  myCurrentBank = bank;
-  uInt16 offset = myCurrentBank << 12;
+  myBankOffset = bank << 12;
 
   // Setup the page access methods for the current bank
-  System::PageAccess access(this, System::PA_READ);
+  System::PageAccess access(this, System::PageAccessType::READ);
 
   // Map Program ROM image into the system
-  for(uInt32 address = 0x1080; address < 0x2000;
-      address += (1 << System::PAGE_SHIFT))
+  for(uInt16 addr = 0x1080; addr < 0x2000; addr += System::PAGE_SIZE)
   {
-    access.codeAccessBase = &myCodeAccessBase[offset + (address & 0x0FFF)];
-    mySystem->setPageAccess(address >> System::PAGE_SHIFT, access);
+    access.romAccessBase = &myRomAccessBase[myBankOffset + (addr & 0x0FFF)];
+    access.romPeekCounter = &myRomAccessCounter[myBankOffset + (addr & 0x0FFF)];
+    access.romPokeCounter = &myRomAccessCounter[myBankOffset + (addr & 0x0FFF) + 24_KB];
+    mySystem->setPageAccess(addr, access);
   }
   return myBankChanged = true;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-uInt16 CartridgeDPCPlus::getBank() const
+uInt16 CartridgeDPCPlus::getBank(uInt16) const
 {
-  return myCurrentBank;
+  return myBankOffset >> 12;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-uInt16 CartridgeDPCPlus::bankCount() const
+uInt16 CartridgeDPCPlus::romBankCount() const
 {
   return 6;
 }
@@ -621,7 +654,7 @@ bool CartridgeDPCPlus::patch(uInt16 address, uInt8 value)
   // For now, we ignore attempts to patch the DPC address space
   if(address >= 0x0080)
   {
-    myProgramImage[(myCurrentBank << 12) + (address & 0x0FFF)] = value;
+    myProgramImage[myBankOffset + (address & 0x0FFF)] = value;
     return myBankChanged = true;
   }
   else
@@ -629,10 +662,19 @@ bool CartridgeDPCPlus::patch(uInt16 address, uInt8 value)
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-const uInt8* CartridgeDPCPlus::getImage(int& size) const
+const ByteBuffer& CartridgeDPCPlus::getImage(size_t& size) const
 {
   size = mySize;
-  return myImage.get();
+  return myImage;
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+uInt8 CartridgeDPCPlus::internalRamGetValue(uInt16 addr) const
+{
+  if(addr < internalRamSize())
+    return myDPCRAM[addr];
+  else
+    return 0;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -640,50 +682,54 @@ bool CartridgeDPCPlus::save(Serializer& out) const
 {
   try
   {
-    out.putString(name());
-
     // Indicates which bank is currently active
-    out.putShort(myCurrentBank);
+    out.putShort(myBankOffset);
 
     // Harmony RAM
-    out.putByteArray(myDPCRAM, 8192);
+    out.putByteArray(myDPCRAM.data(), myDPCRAM.size());
 
     // The top registers for the data fetchers
-    out.putByteArray(myTops, 8);
+    out.putByteArray(myTops.data(), myTops.size());
 
     // The bottom registers for the data fetchers
-    out.putByteArray(myBottoms, 8);
+    out.putByteArray(myBottoms.data(), myBottoms.size());
 
     // The counter registers for the data fetchers
-    out.putShortArray(myCounters, 8);
+    out.putShortArray(myCounters.data(), myCounters.size());
 
     // The counter registers for the fractional data fetchers
-    out.putIntArray(myFractionalCounters, 8);
+    out.putIntArray(myFractionalCounters.data(), myFractionalCounters.size());
 
     // The fractional registers for the data fetchers
-    out.putByteArray(myFractionalIncrements, 8);
+    out.putByteArray(myFractionalIncrements.data(), myFractionalIncrements.size());
 
     // The Fast Fetcher Enabled flag
     out.putBool(myFastFetch);
     out.putBool(myLDAimmediate);
 
     // Control Byte to update
-    out.putByteArray(myParameter, 8);
+    out.putByteArray(myParameter.data(), myParameter.size());
 
     // The music counters
-    out.putIntArray(myMusicCounters, 3);
+    out.putIntArray(myMusicCounters.data(), myMusicCounters.size());
 
     // The music frequencies
-    out.putIntArray(myMusicFrequencies, 3);
+    out.putIntArray(myMusicFrequencies.data(), myMusicFrequencies.size());
 
     // The music waveforms
-    out.putShortArray(myMusicWaveforms, 3);
+    out.putShortArray(myMusicWaveforms.data(), myMusicWaveforms.size());
 
     // The random number generator register
     out.putInt(myRandomNumber);
 
-    out.putInt(mySystemCycles);
-    out.putInt(uInt32(myFractionalClocks * 100000000.0));
+    // Get system cycles and fractional clocks
+    out.putLong(myAudioCycles);
+    out.putDouble(myFractionalClocks);
+
+    // Clock info for Thumbulator
+    out.putLong(myARMCycles);
+
+    CartridgeARM::save(out);
   }
   catch(...)
   {
@@ -699,52 +745,54 @@ bool CartridgeDPCPlus::load(Serializer& in)
 {
   try
   {
-    if(in.getString() != name())
-      return false;
-
     // Indicates which bank is currently active
-    myCurrentBank = in.getShort();
+    myBankOffset = in.getShort();
 
     // Harmony RAM
-    in.getByteArray(myDPCRAM, 8192);
+    in.getByteArray(myDPCRAM.data(), myDPCRAM.size());
 
     // The top registers for the data fetchers
-    in.getByteArray(myTops, 8);
+    in.getByteArray(myTops.data(), myTops.size());
 
     // The bottom registers for the data fetchers
-    in.getByteArray(myBottoms, 8);
+    in.getByteArray(myBottoms.data(), myBottoms.size());
 
     // The counter registers for the data fetchers
-    in.getShortArray(myCounters, 8);
+    in.getShortArray(myCounters.data(), myCounters.size());
 
     // The counter registers for the fractional data fetchers
-    in.getIntArray(myFractionalCounters, 8);
+    in.getIntArray(myFractionalCounters.data(), myFractionalCounters.size());
 
     // The fractional registers for the data fetchers
-    in.getByteArray(myFractionalIncrements, 8);
+    in.getByteArray(myFractionalIncrements.data(), myFractionalIncrements.size());
 
     // The Fast Fetcher Enabled flag
     myFastFetch = in.getBool();
     myLDAimmediate = in.getBool();
 
     // Control Byte to update
-    in.getByteArray(myParameter, 8);
+    in.getByteArray(myParameter.data(), myParameter.size());
 
     // The music mode counters for the data fetchers
-    in.getIntArray(myMusicCounters, 3);
+    in.getIntArray(myMusicCounters.data(), myMusicCounters.size());
 
     // The music mode frequency addends for the data fetchers
-    in.getIntArray(myMusicFrequencies, 3);
+    in.getIntArray(myMusicFrequencies.data(), myMusicFrequencies.size());
 
     // The music waveforms
-    in.getShortArray(myMusicWaveforms, 3);
+    in.getShortArray(myMusicWaveforms.data(), myMusicWaveforms.size());
 
     // The random number generator register
     myRandomNumber = in.getInt();
 
-    // Get system cycles and fractional clocks
-    mySystemCycles = in.getInt();
-    myFractionalClocks = double(in.getInt()) / 100000000.0;
+    // Get audio cycles and fractional clocks
+    myAudioCycles = in.getLong();
+    myFractionalClocks = in.getDouble();
+
+    // Clock info for Thumbulator
+    myARMCycles = in.getLong();
+
+    CartridgeARM::load(in);
   }
   catch(...)
   {
@@ -753,7 +801,7 @@ bool CartridgeDPCPlus::load(Serializer& in)
   }
 
   // Now, go to the current bank
-  bank(myCurrentBank);
+  bank(myBankOffset >> 12);
 
   return true;
 }
