@@ -1,39 +1,36 @@
 //============================================================================
 //
-//   SSSS    tt          lll  lll       
-//  SS  SS   tt           ll   ll        
-//  SS     tttttt  eeee   ll   ll   aaaa 
+//   SSSS    tt          lll  lll
+//  SS  SS   tt           ll   ll
+//  SS     tttttt  eeee   ll   ll   aaaa
 //   SSSS    tt   ee  ee  ll   ll      aa
 //      SS   tt   eeeeee  ll   ll   aaaaa  --  "An Atari 2600 VCS Emulator"
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2016 by Bradford W. Mott, Stephen Anthony
+// Copyright (c) 1995-2022 by Bradford W. Mott, Stephen Anthony
 // and the Stella Team
 //
 // See the file "License.txt" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
-//
-// $Id: MT24LC256.cxx 3316 2016-08-24 23:57:07Z stephena $
 //============================================================================
 
 #include <cstdio>
-#include <fstream>
 
 #include "System.hxx"
 #include "MT24LC256.hxx"
 
-#define DEBUG_EEPROM 0
+//#define DEBUG_EEPROM
 
-#if DEBUG_EEPROM
-  char jpee_msg[256];
-  #define JPEE_LOG0(msg) jpee_logproc(msg)
-  #define JPEE_LOG1(msg,arg1) sprintf(jpee_msg,(msg),(arg1)), jpee_logproc(jpee_msg)
-  #define JPEE_LOG2(msg,arg1,arg2) sprintf(jpee_msg,(msg),(arg1),(arg2)), jpee_logproc(jpee_msg)
+#ifdef DEBUG_EEPROM
+  static char jpee_msg[256];
+  #define JPEE_LOG0(msg) jpee_logproc(msg);
+  #define JPEE_LOG1(msg,arg1) sprintf(jpee_msg,(msg),(arg1)), jpee_logproc(jpee_msg);
+  #define JPEE_LOG2(msg,arg1,arg2) sprintf(jpee_msg,(msg),(arg1),(arg2)), jpee_logproc(jpee_msg);
 #else
-  #define JPEE_LOG0(msg) { }
-  #define JPEE_LOG1(msg,arg1) { }
-  #define JPEE_LOG2(msg,arg1,arg2) { }
+  #define JPEE_LOG0(msg)
+  #define JPEE_LOG1(msg,arg1)
+  #define JPEE_LOG2(msg,arg1,arg2)
 #endif
 
 /*
@@ -46,59 +43,46 @@
 */
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-MT24LC256::MT24LC256(const string& filename, const System& system)
-  : mySystem(system),
-    mySDA(false),
-    mySCL(false),
-    myTimerActive(false),
-    myCyclesWhenTimerSet(0),
-    myCyclesWhenSDASet(0),
-    myCyclesWhenSCLSet(0),
-    myDataFile(filename),
-    myDataFileExists(false),
-    myDataChanged(false),
-    jpee_mdat(0),
-    jpee_sdat(0),
-    jpee_mclk(0),
-    jpee_sizemask(0),
-    jpee_pagemask(0),
-    jpee_smallmode(0),
-    jpee_logmode(0),
-    jpee_pptr(0),
-    jpee_state(0),
-    jpee_nb(0),
-    jpee_address(0),
-    jpee_ad_known(0)
+MT24LC256::MT24LC256(const FilesystemNode& eepromfile, const System& system,
+                     const Controller::onMessageCallback& callback)
+  : mySystem{system},
+    myCallback{callback},
+    myDataFile{eepromfile}
 {
   // Load the data from an external file (if it exists)
-  ifstream in(myDataFile, std::ios_base::binary);
-  if(in.is_open())
+  bool fileValid = false;
+  try
   {
-    // Get length of file; it must be 32768
-    in.seekg(0, std::ios::end);
-    if(uInt32(in.tellg()) == 32768u)
-    {
-      in.seekg(0, std::ios::beg);
-      in.read(reinterpret_cast<char*>(myData), 32768);
-      myDataFileExists = true;
-    }
+    // A valid file must be 32768 bytes; otherwise we create a new one
+    if(myDataFile.read(myData) == FLASH_SIZE)
+      fileValid = true;
   }
-  else
-    myDataFileExists = false;
+  catch(...)
+  {
+    fileValid = false;
+  }
+
+  if(!fileValid)
+  {
+    myData = make_unique<uInt8[]>(FLASH_SIZE);
+    std::fill_n(myData.get(), FLASH_SIZE, INITIAL_VALUE);
+    myDataChanged = true;
+   }
 
   // Then initialize the I2C state
   jpee_init();
+
+  systemReset();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 MT24LC256::~MT24LC256()
 {
   // Save EEPROM data to external file only when necessary
-  if(!myDataFileExists || myDataChanged)
+  if(myDataChanged)
   {
-    ofstream out(myDataFile, std::ios_base::binary);
-    if(out.is_open())
-      out.write(reinterpret_cast<char*>(myData), 32768);
+    try { myDataFile.write(myData, FLASH_SIZE); }
+    catch(...) { }
   }
 }
 
@@ -137,7 +121,7 @@ void MT24LC256::update()
   // we only do the write when they have the same 'timestamp'
   if(myCyclesWhenSDASet == myCyclesWhenSCLSet)
   {
-#if DEBUG_EEPROM
+#ifdef DEBUG_EEPROM
     cerr << endl << "  I2C_PIN_WRITE(SCL = " << mySCL
          << ", SDA = " << mySDA << ")" << " @ " << mySystem.cycles() << endl;
 #endif
@@ -147,21 +131,38 @@ void MT24LC256::update()
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void MT24LC256::erase()
+void MT24LC256::systemReset()
 {
-  memset(myData, 0xff, 32768);
+  myPageHit.fill(false);
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+void MT24LC256::eraseAll()
+{
+  std::fill_n(myData.get(), FLASH_SIZE, INITIAL_VALUE);
   myDataChanged = true;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void MT24LC256::systemCyclesReset()
+void MT24LC256::eraseCurrent()
 {
-  // System cycles are being reset to zero so we need to adjust
-  // the cycle counts we remembered
-  uInt32 cycles = mySystem.cycles();
-  myCyclesWhenSDASet -= cycles;
-  myCyclesWhenSCLSet -= cycles;
-  myCyclesWhenTimerSet -= cycles;
+  for(uInt32 page = 0; page < PAGE_NUM; ++page)
+  {
+    if(myPageHit[page])
+    {
+      std::fill_n(myData.get() + page * PAGE_SIZE, PAGE_SIZE, INITIAL_VALUE);
+      myDataChanged = true;
+    }
+  }
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+bool MT24LC256::isPageUsed(uInt32 page) const
+{
+  if(page < PAGE_NUM)
+    return myPageHit[page];
+  else
+    return false;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -170,12 +171,10 @@ void MT24LC256::jpee_init()
   jpee_sdat = 1;
   jpee_address = 0;
   jpee_state=0;
-  jpee_sizemask = 32767;
-  jpee_pagemask = 63;
+  jpee_sizemask = FLASH_SIZE - 1;
+  jpee_pagemask = PAGE_SIZE - 1;
   jpee_smallmode = 0;
   jpee_logmode = -1;
-  if(!myDataFileExists)
-    memset(myData, 0xff, 32768);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -184,21 +183,21 @@ void MT24LC256::jpee_data_start()
   /* We have a start condition */
   if (jpee_state == 1 && (jpee_nb != 1 || jpee_pptr != 3))
   {
-    JPEE_LOG0("I2C_WARNING ABANDON WRITE");
+    JPEE_LOG0("I2C_WARNING ABANDON WRITE")
     jpee_ad_known = 0;
   }
   if (jpee_state == 3)
   {
-    JPEE_LOG0("I2C_WARNING ABANDON READ");
+    JPEE_LOG0("I2C_WARNING ABANDON READ")
   }
   if (!jpee_timercheck(0))
   {
-    JPEE_LOG0("I2C_START");
+    JPEE_LOG0("I2C_START")
     jpee_state = 2;
   }
   else
   {
-    JPEE_LOG0("I2C_BUSY");
+    JPEE_LOG0("I2C_BUSY")
     jpee_state = 0;
   }
   jpee_pptr = 0;
@@ -211,35 +210,41 @@ void MT24LC256::jpee_data_stop()
 {
   if (jpee_state == 1 && jpee_nb != 1)
   {
-    JPEE_LOG0("I2C_WARNING ABANDON_WRITE");
+    JPEE_LOG0("I2C_WARNING ABANDON_WRITE")
     jpee_ad_known = 0;
   }
   if (jpee_state == 3)
   {
-    JPEE_LOG0("I2C_WARNING ABANDON_READ");
+    JPEE_LOG0("I2C_WARNING ABANDON_READ")
     jpee_ad_known = 0;
   }
   /* We have a stop condition. */
   if (jpee_state == 1 && jpee_nb == 1 && jpee_pptr > 3)
   {
     jpee_timercheck(1);
-    JPEE_LOG2("I2C_STOP(Write %d bytes at %04X)",jpee_pptr-3,jpee_address);
+    JPEE_LOG2("I2C_STOP(Write %d bytes at %04X)",jpee_pptr-3,jpee_address)
     if (((jpee_address + jpee_pptr-4) ^ jpee_address) & ~jpee_pagemask)
     {
       jpee_pptr = 4+jpee_pagemask-(jpee_address & jpee_pagemask);
-      JPEE_LOG1("I2C_WARNING PAGECROSSING!(Truncate to %d bytes)",jpee_pptr-3);
+      JPEE_LOG1("I2C_WARNING PAGECROSSING!(Truncate to %d bytes)",jpee_pptr-3)
     }
     for (int i=3; i<jpee_pptr; i++)
     {
       myDataChanged = true;
+      myPageHit[jpee_address / PAGE_SIZE] = true;
+
+      myCallback("AtariVox/SaveKey EEPROM write");
+
       myData[(jpee_address++) & jpee_sizemask] = jpee_packet[i];
       if (!(jpee_address & jpee_pagemask))
         break;  /* Writes can't cross page boundary! */
     }
     jpee_ad_known = 0;
   }
+#ifdef DEBUG_EEPROM
   else
-    JPEE_LOG0("I2C_STOP");
+    jpee_logproc("I2C_STOP");
+#endif
 
   jpee_state = 0;
 }
@@ -260,28 +265,31 @@ void MT24LC256::jpee_clock_fall()
           if (jpee_smallmode && ((jpee_nb & 0xF0) == 0xA0))
           {
             jpee_packet[1] = (jpee_nb >> 1) & 7;
+          #ifdef DEBUG_EEPROM
             if (jpee_packet[1] != (jpee_address >> 8) && (jpee_packet[0] & 1))
-              JPEE_LOG0("I2C_WARNING ADDRESS MSB CHANGED");
+              jpee_logproc("I2C_WARNING ADDRESS MSB CHANGED");
+          #endif
             jpee_nb &= 0x1A1;
           }
           if (jpee_nb == 0x1A0)
           {
-            JPEE_LOG1("I2C_SENT(%02X--start write)",jpee_packet[0]);
+            JPEE_LOG1("I2C_SENT(%02X--start write)",jpee_packet[0])
             jpee_state = 2;
             jpee_sdat = 0;
           }
           else if (jpee_nb == 0x1A1)
           {
             jpee_state = 4;
-            JPEE_LOG2("I2C_SENT(%02X--start read @%04X)",
-            jpee_packet[0],jpee_address);
+            JPEE_LOG2("I2C_SENT(%02X--start read @%04X)", jpee_packet[0],jpee_address)
+          #ifdef DEBUG_EEPROM
             if (!jpee_ad_known)
-              JPEE_LOG0("I2C_WARNING ADDRESS IS UNKNOWN");
+              jpee_logproc("I2C_WARNING ADDRESS IS UNKNOWN");
+          #endif
             jpee_sdat = 0;
           }
           else
           {
-            JPEE_LOG1("I2C_WARNING ODDBALL FIRST BYTE!(%02X)",jpee_nb & 0xFF);
+            JPEE_LOG1("I2C_WARNING ODDBALL FIRST BYTE!(%02X)",jpee_nb & 0xFF)
             jpee_state = 0;
           }
         }
@@ -306,14 +314,16 @@ void MT24LC256::jpee_clock_fall()
         }
         else if (jpee_pptr < 70)
         {
-          JPEE_LOG1("I2C_SENT(%02X)",jpee_nb & 0xFF);
+          JPEE_LOG1("I2C_SENT(%02X)",jpee_nb & 0xFF)
           jpee_packet[jpee_pptr++] = uInt8(jpee_nb);
           jpee_address = (jpee_packet[1] << 8) | jpee_packet[2];
           if (jpee_pptr > 2)
             jpee_ad_known = 1;
         }
+      #ifdef DEBUG_EEPROM
         else
-          JPEE_LOG0("I2C_WARNING OUTPUT_OVERFLOW!");
+          jpee_logproc("I2C_WARNING OUTPUT_OVERFLOW!");
+      #endif
       }
       jpee_sdat = 1;
       jpee_nb = 1;
@@ -323,13 +333,18 @@ void MT24LC256::jpee_clock_fall()
     case 4:
       if (jpee_mdat && jpee_sdat)
       {
-        JPEE_LOG0("I2C_READ_NAK");
+        JPEE_LOG0("I2C_READ_NAK")
         jpee_state=0;
         break;
       }
       jpee_state=3;
+      myPageHit[jpee_address / PAGE_SIZE] = true;
+
+      myCallback("AtariVox/SaveKey EEPROM read");
+
       jpee_nb = (myData[jpee_address & jpee_sizemask] << 1) | 1;  /* Fall through */
-      JPEE_LOG2("I2C_READ(%04X=%02X)",jpee_address,jpee_nb/2);
+      JPEE_LOG2("I2C_READ(%04X=%02X)",jpee_address,jpee_nb/2)
+      [[fallthrough]];
 
     case 3:
       jpee_sdat = !!(jpee_nb & 256);
@@ -338,7 +353,7 @@ void MT24LC256::jpee_clock_fall()
       {
         jpee_state = 4;
         jpee_sdat = 1;
-        jpee_address++;
+        ++jpee_address;
       }
       break;
 
@@ -346,7 +361,7 @@ void MT24LC256::jpee_clock_fall()
       /* Do nothing */
       break;
   }
-  JPEE_LOG2("I2C_CLOCK (dat=%d/%d)",jpee_mdat,jpee_sdat);
+  JPEE_LOG2("I2C_CLOCK (dat=%d/%d)",jpee_mdat,jpee_sdat)
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -367,16 +382,9 @@ bool MT24LC256::jpee_timercheck(int mode)
   {
     if(myTimerActive)
     {
-      uInt32 elapsed = mySystem.cycles() - myCyclesWhenTimerSet;
-      myTimerActive = elapsed < uInt32(5000000.0 / 838.0);
+      const uInt64 elapsed = mySystem.cycles() - myCyclesWhenTimerSet;
+      myTimerActive = elapsed < static_cast<uInt64>(5000000.0 / 838.0);
     }
     return myTimerActive;
   }
 }
-
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-int MT24LC256::jpee_logproc(char const *st)
-{
-  cerr << "    " << st << endl;
-  return 0;
-}
