@@ -1,20 +1,18 @@
 //============================================================================
 //
-//   SSSS    tt          lll  lll       
-//  SS  SS   tt           ll   ll        
-//  SS     tttttt  eeee   ll   ll   aaaa 
+//   SSSS    tt          lll  lll
+//  SS  SS   tt           ll   ll
+//  SS     tttttt  eeee   ll   ll   aaaa
 //   SSSS    tt   ee  ee  ll   ll      aa
 //      SS   tt   eeeeee  ll   ll   aaaaa  --  "An Atari 2600 VCS Emulator"
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2016 by Bradford W. Mott, Stephen Anthony
+// Copyright (c) 1995-2022 by Bradford W. Mott, Stephen Anthony
 // and the Stella Team
 //
 // See the file "License.txt" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
-//
-// $Id: Launcher.cxx 3304 2016-04-03 00:35:00Z stephena $
 //============================================================================
 
 #include "LauncherDialog.hxx"
@@ -29,30 +27,43 @@
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 Launcher::Launcher(OSystem& osystem)
-  : DialogContainer(osystem)
+  : DialogContainer(osystem),
+    mySize{myOSystem.settings().getSize("launcherres")}
 {
-  const GUI::Size& s = myOSystem.settings().getSize("launcherres");
-  const GUI::Size& d = myOSystem.frameBuffer().desktopSize();
-  myWidth = s.w;  myHeight = s.h;
+  const Common::Size& d = myOSystem.frameBuffer().desktopSize(BufferType::Launcher);
+  double overscan = 1 - myOSystem.settings().getInt("tia.fs_overscan") / 100.0;
 
   // The launcher dialog is resizable, within certain bounds
   // We check those bounds now
-  myWidth  = std::max(myWidth,  uInt32(FrameBuffer::kFBMinW));
-  myHeight = std::max(myHeight, uInt32(FrameBuffer::kFBMinH));
-  myWidth  = std::min(myWidth,  uInt32(d.w));
-  myHeight = std::min(myHeight, uInt32(d.h));
+  mySize.clamp(FBMinimum::Width, d.w, FBMinimum::Height, d.h);
+  // Do not include overscan when launcher saving size
+  myOSystem.settings().setValue("launcherres", mySize);
+  // Now make overscan effective
+  mySize.w = std::min(mySize.w, uInt32(d.w * overscan));
+  mySize.h = std::min(mySize.h, uInt32(d.h * overscan));
 
-  myOSystem.settings().setValue("launcherres",
-                                GUI::Size(myWidth, myHeight));
+  myBaseDialog = new LauncherDialog(myOSystem, *this, 0, 0, mySize.w, mySize.h);
+}
 
-  myBaseDialog = new LauncherDialog(myOSystem, *this, 0, 0, myWidth, myHeight);
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Launcher::~Launcher()
+{
+  delete myBaseDialog;  myBaseDialog = nullptr;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 FBInitStatus Launcher::initializeVideo()
 {
   string title = string("Stella ") + STELLA_VERSION;
-  return myOSystem.frameBuffer().createDisplay(title, myWidth, myHeight);
+  return myOSystem.frameBuffer().createDisplay(
+      title, BufferType::Launcher, mySize
+  );
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+const string& Launcher::selectedRom()
+{
+  return (static_cast<LauncherDialog*>(myBaseDialog))->selectedRom();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -62,9 +73,9 @@ const string& Launcher::selectedRomMD5()
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-const FilesystemNode& Launcher::currentNode() const
+const FilesystemNode& Launcher::currentDir() const
 {
-  return (static_cast<LauncherDialog*>(myBaseDialog))->currentNode();
+  return (static_cast<LauncherDialog*>(myBaseDialog))->currentDir();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -72,3 +83,15 @@ void Launcher::reload()
 {
   (static_cast<LauncherDialog*>(myBaseDialog))->reload();
 }
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+void Launcher::quit()
+{
+  (static_cast<LauncherDialog*>(myBaseDialog))->quit();
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+Dialog* Launcher::baseDialog()
+{
+  return myBaseDialog;
+}
