@@ -1,73 +1,79 @@
 //============================================================================
 //
-//   SSSS    tt          lll  lll       
-//  SS  SS   tt           ll   ll        
-//  SS     tttttt  eeee   ll   ll   aaaa 
+//   SSSS    tt          lll  lll
+//  SS  SS   tt           ll   ll
+//  SS     tttttt  eeee   ll   ll   aaaa
 //   SSSS    tt   ee  ee  ll   ll      aa
 //      SS   tt   eeeeee  ll   ll   aaaaa  --  "An Atari 2600 VCS Emulator"
 //  SS  SS   tt   ee      ll   ll  aa  aa
 //   SSSS     ttt  eeeee llll llll  aaaaa
 //
-// Copyright (c) 1995-2016 by Bradford W. Mott, Stephen Anthony
+// Copyright (c) 1995-2022 by Bradford W. Mott, Stephen Anthony
 // and the Stella Team
 //
 // See the file "License.txt" for information on usage and redistribution of
 // this file, and for a DISCLAIMER OF ALL WARRANTIES.
-//
-// $Id: ListWidget.cxx 3302 2016-04-02 23:47:46Z stephena $
 //============================================================================
 
-#include <cctype>
-#include <algorithm>
-
 #include "OSystem.hxx"
 #include "Widget.hxx"
 #include "ScrollBarWidget.hxx"
 #include "Dialog.hxx"
 #include "FrameBuffer.hxx"
+#include "StellaKeys.hxx"
+#include "EventHandler.hxx"
 #include "ListWidget.hxx"
-#include "bspf.hxx"
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 ListWidget::ListWidget(GuiObject* boss, const GUI::Font& font,
-                       int x, int y, int w, int h, bool quickSelect)
+                       int x, int y, int w, int h, bool useScrollbar)
   : EditableWidget(boss, font, x, y, 16, 16),
-    _rows(0),
-    _cols(0),
-    _currentPos(0),
-    _selectedItem(-1),
-    _highlightedItem(-1),
-    _editMode(false),
-    _currentKeyDown(KBDK_UNKNOWN),
-    _quickSelect(quickSelect),
-    _quickSelectTime(0)
+    _useScrollbar{useScrollbar}
 {
-  _flags = WIDGET_ENABLED | WIDGET_CLEARBG | WIDGET_RETAIN_FOCUS;
+  _flags = Widget::FLAG_ENABLED | Widget::FLAG_CLEARBG | Widget::FLAG_RETAIN_FOCUS;
   _bgcolor = kWidColor;
   _bgcolorhi = kWidColor;
   _textcolor = kTextColor;
   _textcolorhi = kTextColor;
 
+  _editMode = false;
+
   _cols = w / _fontWidth;
-  _rows = h / _fontHeight;
+  _rows = h / _lineHeight;
 
   // Set real dimensions
-  _w = w - kScrollBarWidth;
   _h = h + 2;
 
   // Create scrollbar and attach to the list
-  _scrollBar = new ScrollBarWidget(boss, font, _x + _w, _y, kScrollBarWidth, _h);
-  _scrollBar->setTarget(this);
+  if(_useScrollbar)
+  {
+    _w = w - ScrollBarWidget::scrollBarWidth(_font);
+    _scrollBar = new ScrollBarWidget(boss, font, _x + _w, _y,
+                                     ScrollBarWidget::scrollBarWidth(_font), _h);
+    _scrollBar->setTarget(this);
+  }
+  else
+    _w = w - 1;
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+void ListWidget::setHeight(int h)
+{
+  Widget::setHeight(h);
+  if(_useScrollbar)
+    _scrollBar->setHeight(h);
+
+  _rows = (h - 2) / _lineHeight;
+  recalc();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void ListWidget::setSelected(int item)
 {
-  if(item < 0 || item >= int(_list.size()))
-  {
-    setDirty();  // Simply redraw and exit
+  setDirty();
+
+  if(item < 0 || item >= static_cast<int>(_list.size()))
     return;
-  }
 
   if(isEnabled())
   {
@@ -112,7 +118,7 @@ void ListWidget::setSelected(const string& item)
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void ListWidget::setHighlighted(int item)
 {
-  if(item < -1 || item >= int(_list.size()))
+  if(item < -1 || item >= static_cast<int>(_list.size()))
     return;
 
   if(isEnabled())
@@ -135,14 +141,14 @@ void ListWidget::setHighlighted(int item)
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 const string& ListWidget::getSelectedString() const
 {
-  return (_selectedItem >= 0 && _selectedItem < int(_list.size()))
+  return (_selectedItem >= 0 && _selectedItem < static_cast<int>(_list.size()))
             ? _list[_selectedItem] : EmptyString;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void ListWidget::scrollTo(int item)
 {
-  int size = int(_list.size());
+  const int size = static_cast<int>(_list.size());
   if (item >= size)
     item = size - 1;
   if (item < 0)
@@ -155,46 +161,70 @@ void ListWidget::scrollTo(int item)
   }
 }
 
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+int ListWidget::getWidth() const
+{
+  return _w + ScrollBarWidget::scrollBarWidth(_font);
+}
+
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void ListWidget::recalc()
 {
-  int size = int(_list.size());
+  const int size = static_cast<int>(_list.size());
 
-  if (_currentPos >= size)
-    _currentPos = size - 1;
+  if(_currentPos >= size - _rows)
+  {
+    if(size <= _rows)
+      _currentPos = 0;
+    else
+      _currentPos = size - _rows;
+  }
   if (_currentPos < 0)
     _currentPos = 0;
 
-  if(_selectedItem < 0 || _selectedItem >= size)
+  if(_selectedItem >= size)
+    _selectedItem = size - 1;
+  if(_selectedItem < 0)
     _selectedItem = 0;
 
   _editMode = false;
 
-  _scrollBar->_numEntries     = int(_list.size());
-  _scrollBar->_entriesPerPage = _rows;
+  if(_useScrollbar)
+  {
+    _scrollBar->_numEntries = static_cast<int>(_list.size());
+    _scrollBar->_entriesPerPage = _rows;
+    // disable scrollbar if no longer necessary
+    scrollBarRecalc();
+  }
 
   // Reset to normal data entry
   abortEditMode();
+
+  setDirty();
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void ListWidget::scrollBarRecalc()
 {
-  _scrollBar->_currentPos = _currentPos;
-  _scrollBar->recalc();
-  sendCommand(ListWidget::kScrolledCmd, _currentPos, _id);
+  if(_useScrollbar)
+  {
+    _scrollBar->_currentPos = _currentPos;
+    _scrollBar->recalc();
+    sendCommand(ListWidget::kScrolledCmd, _currentPos, _id);
+  }
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void ListWidget::handleMouseDown(int x, int y, int button, int clickCount)
+void ListWidget::handleMouseDown(int x, int y, MouseButton b, int clickCount)
 {
   if (!isEnabled())
     return;
 
+  resetSelection();
   // First check whether the selection changed
   int newSelectedItem;
   newSelectedItem = findItem(x, y);
-  if (newSelectedItem >= int(_list.size()))
+  if (newSelectedItem >= static_cast<int>(_list.size()))
     return;
 
   if (_selectedItem != newSelectedItem)
@@ -205,13 +235,13 @@ void ListWidget::handleMouseDown(int x, int y, int button, int clickCount)
     sendCommand(ListWidget::kSelectionChangedCmd, _selectedItem, _id);
     setDirty();
   }
-	
+
   // TODO: Determine where inside the string the user clicked and place the
   // caret accordingly. See _editScrollOffset and EditTextWidget::handleMouseDown.
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-void ListWidget::handleMouseUp(int x, int y, int button, int clickCount)
+void ListWidget::handleMouseUp(int x, int y, MouseButton b, int clickCount)
 {
   // If this was a double click and the mouse is still over the selected item,
   // send the double click command
@@ -228,73 +258,29 @@ void ListWidget::handleMouseUp(int x, int y, int button, int clickCount)
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void ListWidget::handleMouseWheel(int x, int y, int direction)
 {
-  _scrollBar->handleMouseWheel(x, y, direction);
+  if(_useScrollbar)
+    _scrollBar->handleMouseWheel(x, y, direction);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 int ListWidget::findItem(int x, int y) const
 {
-  return (y - 1) / _fontHeight + _currentPos;
+  return (y - 1) / _lineHeight + _currentPos;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 bool ListWidget::handleText(char text)
 {
-  bool handled = true;
-  int oldSelectedItem = _selectedItem;
-
-  if (!_editMode && _quickSelect)
-  {
-    // Quick selection mode: Go to first list item starting with this key
-    // (or a substring accumulated from the last couple key presses).
-    // Only works in a useful fashion if the list entries are sorted.
-    // TODO: Maybe this should be off by default, and instead we add a
-    // method "enableQuickSelect()" or so ?
-    uInt64 time = instance().getTicks() / 1000;
-    if (_quickSelectTime < time)
-      _quickSelectStr = text;
-    else
-      _quickSelectStr += text;
-    _quickSelectTime = time + _QUICK_SELECT_DELAY;
-
-    // FIXME: This is bad slow code (it scans the list linearly each time a
-    // key is pressed); it could be much faster. Only of importance if we have
-    // quite big lists to deal with -- so for now we can live with this lazy
-    // implementation :-)
-    int newSelectedItem = 0;
-    for(const auto& i: _list)
-    {
-      if(BSPF::startsWithIgnoreCase(i, _quickSelectStr))
-      {
-        _selectedItem = newSelectedItem;
-        break;
-      }
-      newSelectedItem++;
-    }
-  }
-  else if (_editMode)
-  {
-    // Class EditableWidget handles all text editing related key presses for us
-    handled = EditableWidget::handleText(text);
-  }
-
-  if (_selectedItem != oldSelectedItem)
-  {
-    _scrollBar->draw();
-    scrollToSelected();
-
-    sendCommand(ListWidget::kSelectionChangedCmd, _selectedItem, _id);
-  }
-
-  return handled;
+  // Class EditableWidget handles all text editing related key presses for us
+  return _editMode ? EditableWidget::handleText(text) : true;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 bool ListWidget::handleKeyDown(StellaKey key, StellaMod mod)
 {
   // Ignore all Alt-mod keys
-  if(instance().eventHandler().kbdAlt(mod))
-    return true;
+  if(StellaModTest::isAlt(mod))
+    return false;
 
   bool handled = true;
   if (!_editMode)
@@ -315,16 +301,22 @@ bool ListWidget::handleKeyDown(StellaKey key, StellaMod mod)
     }
   }
 
-  _currentKeyDown = key;
   return handled;
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-bool ListWidget::handleKeyUp(StellaKey key, StellaMod mod)
+void ListWidget::handleJoyDown(int stick, int button, bool longPress)
 {
-  if (key == _currentKeyDown)
-    _currentKeyDown = KBDK_UNKNOWN;
-  return true;
+  if (longPress)
+    sendCommand(ListWidget::kLongButtonPressCmd, _selectedItem, _id);
+}
+
+// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
+void ListWidget::handleJoyUp(int stick, int button)
+{
+  const Event::Type e = _boss->instance().eventHandler().eventForJoyButton(EventMode::kMenuMode, stick, button);
+
+  handleEvent(e);
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -334,8 +326,8 @@ bool ListWidget::handleEvent(Event::Type e)
     return false;
 
   bool handled = true;
-  int oldSelectedItem = _selectedItem;
-  int size = int(_list.size());
+  const int oldSelectedItem = _selectedItem;
+  const int size = static_cast<int>(_list.size());
 
   switch(e)
   {
@@ -360,12 +352,14 @@ bool ListWidget::handleEvent(Event::Type e)
       break;
 
     case Event::UIPgUp:
+    case Event::UILeft:
       _selectedItem -= _rows - 1;
       if (_selectedItem < 0)
         _selectedItem = 0;
       break;
 
     case Event::UIPgDown:
+    case Event::UIRight:
       _selectedItem += _rows - 1;
       if (_selectedItem >= size)
         _selectedItem = size - 1;
@@ -380,7 +374,7 @@ bool ListWidget::handleEvent(Event::Type e)
       break;
 
     case Event::UIPrevDir:
-      sendCommand(ListWidget::kPrevDirCmd, _selectedItem, _id);
+      sendCommand(ListWidget::kParentDirCmd, _selectedItem, _id);
       break;
 
     default:
@@ -389,8 +383,11 @@ bool ListWidget::handleEvent(Event::Type e)
 
   if (_selectedItem != oldSelectedItem)
   {
-    _scrollBar->draw();
-    scrollToSelected();
+    if(_useScrollbar)
+    {
+      _scrollBar->draw();
+      scrollToSelected();
+    }
 
     sendCommand(ListWidget::kSelectionChangedCmd, _selectedItem, _id);
   }
@@ -410,18 +407,16 @@ void ListWidget::lostFocusWidget()
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 void ListWidget::handleCommand(CommandSender* sender, int cmd, int data, int id)
 {
-  switch (cmd)
+  if (cmd == GuiObject::kSetPositionCmd)
   {
-    case kSetPositionCmd:
-      if (_currentPos != data)
-      {
-        _currentPos = data;
-        setDirty();
+    if (_currentPos != data)
+    {
+      _currentPos = data;
+      setDirty();
 
-        // Let boss know the list has scrolled
-        sendCommand(ListWidget::kScrolledCmd, _currentPos, _id);
-      }
-      break;
+      // Let boss know the list has scrolled
+      sendCommand(ListWidget::kScrolledCmd, _currentPos, _id);
+    }
   }
 }
 
@@ -440,19 +435,22 @@ void ListWidget::scrollToCurrent(int item)
     _currentPos = item - _rows + 1;
   }
 
-  if (_currentPos < 0 || _rows > int(_list.size()))
+  if (_currentPos < 0 || _rows > static_cast<int>(_list.size()))
     _currentPos = 0;
-  else if (_currentPos + _rows > int(_list.size()))
-    _currentPos = int(_list.size()) - _rows;
+  else if (_currentPos + _rows > static_cast<int>(_list.size()))
+    _currentPos = static_cast<int>(_list.size()) - _rows;
 
-  int oldScrollPos = _scrollBar->_currentPos;
-  _scrollBar->_currentPos = _currentPos;
-  _scrollBar->recalc();
+  if(_useScrollbar)
+  {
+    const int oldScrollPos = _scrollBar->_currentPos;
+    _scrollBar->_currentPos = _currentPos;
+    _scrollBar->recalc();
 
-  setDirty();
+    setDirty();
 
-  if(oldScrollPos != _currentPos)
-    sendCommand(ListWidget::kScrolledCmd, _currentPos, _id);
+    if(oldScrollPos != _currentPos)
+      sendCommand(ListWidget::kScrolledCmd, _currentPos, _id);
+  }
 }
 
 // - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
@@ -492,6 +490,3 @@ void ListWidget::abortEditMode()
   // Reset to normal data entry
   EditableWidget::abortEditMode();
 }
-
-// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
-uInt64 ListWidget::_QUICK_SELECT_DELAY = 300;
